# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/queries.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from chalk._gen.chalk.chart.v1 import densetimeserieschart_pb2 as chalk_dot_chart_dot_v1_dot_densetimeserieschart__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1d\x63halk/server/v1/queries.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a)chalk/chart/v1/densetimeserieschart.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto"F\n!GetQueryPerformanceSummaryRequest\x12!\n\x0coperation_id\x18\x01 \x01(\tR\x0boperationId"\x95\x01\n"GetQueryPerformanceSummaryResponse\x12!\n\x0coperation_id\x18\x01 \x01(\tR\x0boperationId\x12\x34\n\x13performance_summary\x18\x02 \x01(\tH\x00R\x12performanceSummary\x88\x01\x01\x42\x16\n\x14_performance_summary"\x8c\x01\n\x18ListQueryErrorsPageToken\x12$\n\x0enumeric_id_hwm\x18\x01 \x01(\x03R\x0cnumericIdHwm\x12J\n\x13\x65rror_timestamp_hwm\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x11\x65rrorTimestampHwm"\xee\x01\n\x11QueryErrorFilters\x12&\n\x0coperation_id\x18\x01 \x01(\tH\x00R\x0boperationId\x88\x01\x01\x12$\n\x0b\x66\x65\x61ture_fqn\x18\x02 \x01(\tH\x01R\nfeatureFqn\x88\x01\x01\x12&\n\x0cresolver_fqn\x18\x03 \x01(\tH\x02R\x0bresolverFqn\x88\x01\x01\x12"\n\nquery_name\x18\x04 \x01(\tH\x03R\tqueryName\x88\x01\x01\x42\x0f\n\r_operation_idB\x0e\n\x0c_feature_fqnB\x0f\n\r_resolver_fqnB\r\n\x0b_query_name"\x8e\x07\n\x0eQueryErrorMeta\x12\x0e\n\x02id\x18\x01 \x01(\x03R\x02id\x12\x12\n\x04\x63ode\x18\x02 \x01(\tR\x04\x63ode\x12\x1a\n\x08\x63\x61tegory\x18\x03 \x01(\tR\x08\x63\x61tegory\x12\x18\n\x07message\x18\x04 \x01(\tR\x07message\x12\x33\n\x13\x64isplay_primary_key\x18\x05 \x01(\tH\x00R\x11\x64isplayPrimaryKey\x88\x01\x01\x12:\n\x17\x64isplay_primary_key_fqn\x18\x06 \x01(\tH\x01R\x14\x64isplayPrimaryKeyFqn\x88\x01\x01\x12\x1d\n\x07\x66\x65\x61ture\x18\x07 \x01(\tH\x02R\x07\x66\x65\x61ture\x88\x01\x01\x12\x1f\n\x08resolver\x18\x08 \x01(\tH\x03R\x08resolver\x88\x01\x01\x12"\n\nquery_name\x18\t \x01(\tH\x04R\tqueryName\x88\x01\x01\x12*\n\x0e\x65xception_kind\x18\n \x01(\tH\x05R\rexceptionKind\x88\x01\x01\x12\x30\n\x11\x65xception_message\x18\x0b \x01(\tH\x06R\x10\x65xceptionMessage\x88\x01\x01\x12\x36\n\x14\x65xception_stacktrace\x18\x0c \x01(\tH\x07R\x13\x65xceptionStacktrace\x88\x01\x01\x12G\n\x1d\x65xception_internal_stacktrace\x18\r \x01(\tH\x08R\x1b\x65xceptionInternalStacktrace\x88\x01\x01\x12&\n\x0coperation_id\x18\x0e \x01(\tH\tR\x0boperationId\x88\x01\x01\x12(\n\rdeployment_id\x18\x0f \x01(\tH\nR\x0c\x64\x65ploymentId\x88\x01\x01\x12\x39\n\ncreated_at\x18\x10 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAtB\x16\n\x14_display_primary_keyB\x1a\n\x18_display_primary_key_fqnB\n\n\x08_featureB\x0b\n\t_resolverB\r\n\x0b_query_nameB\x11\n\x0f_exception_kindB\x14\n\x12_exception_messageB\x17\n\x15_exception_stacktraceB \n\x1e_exception_internal_stacktraceB\x0f\n\r_operation_idB\x10\n\x0e_deployment_id"\xab\x02\n\x16ListQueryErrorsRequest\x12\x39\n\nstart_date\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tstartDate\x12\x35\n\x08\x65nd_date\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x07\x65ndDate\x12<\n\x07\x66ilters\x18\x03 \x01(\x0b\x32".chalk.server.v1.QueryErrorFiltersR\x07\x66ilters\x12 \n\tpage_size\x18\x04 \x01(\x05H\x00R\x08pageSize\x88\x01\x01\x12"\n\npage_token\x18\x05 \x01(\tH\x01R\tpageToken\x88\x01\x01\x42\x0c\n\n_page_sizeB\r\n\x0b_page_token"\x9e\x01\n\x17ListQueryErrorsResponse\x12\x42\n\x0cquery_errors\x18\x01 \x03(\x0b\x32\x1f.chalk.server.v1.QueryErrorMetaR\x0bqueryErrors\x12+\n\x0fnext_page_token\x18\x02 \x01(\tH\x00R\rnextPageToken\x88\x01\x01\x42\x12\n\x10_next_page_token"\xe7\x02\n\x1aGetQueryErrorsChartRequest\x12V\n\x19start_timestamp_inclusive\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x17startTimestampInclusive\x12W\n\x17\x65nd_timestamp_exclusive\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00R\x15\x65ndTimestampExclusive\x88\x01\x01\x12>\n\rwindow_period\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0cwindowPeriod\x12<\n\x07\x66ilters\x18\x04 \x01(\x0b\x32".chalk.server.v1.QueryErrorFiltersR\x07\x66iltersB\x1a\n\x18_end_timestamp_exclusive"Y\n\x1bGetQueryErrorsChartResponse\x12:\n\x05\x63hart\x18\x01 \x01(\x0b\x32$.chalk.chart.v1.DenseTimeSeriesChartR\x05\x63hart"9\n\x13GetQueryPlanRequest\x12"\n\rquery_plan_id\x18\x01 \x01(\tR\x0bqueryPlanId"\xc1\x01\n\tQueryPlan\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12%\n\x0e\x65nvironment_id\x18\x02 \x01(\tR\renvironmentId\x12#\n\rdeployment_id\x18\x03 \x01(\tR\x0c\x64\x65ploymentId\x12\x1d\n\nquery_plan\x18\x04 \x01(\tR\tqueryPlan\x12\x39\n\ncreated_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt"Q\n\x14GetQueryPlanResponse\x12\x39\n\nquery_plan\x18\x01 \x01(\x0b\x32\x1a.chalk.server.v1.QueryPlanR\tqueryPlan2\xe1\x03\n\x0eQueriesService\x12\x8a\x01\n\x1aGetQueryPerformanceSummary\x12\x32.chalk.server.v1.GetQueryPerformanceSummaryRequest\x1a\x33.chalk.server.v1.GetQueryPerformanceSummaryResponse"\x03\x80}\x06\x12i\n\x0fListQueryErrors\x12\'.chalk.server.v1.ListQueryErrorsRequest\x1a(.chalk.server.v1.ListQueryErrorsResponse"\x03\x80}\x06\x12u\n\x13GetQueryErrorsChart\x12+.chalk.server.v1.GetQueryErrorsChartRequest\x1a,.chalk.server.v1.GetQueryErrorsChartResponse"\x03\x80}\x06\x12`\n\x0cGetQueryPlan\x12$.chalk.server.v1.GetQueryPlanRequest\x1a%.chalk.server.v1.GetQueryPlanResponse"\x03\x80}\x06\x42\x95\x01\n\x13\x63om.chalk.server.v1B\x0cQueriesProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.queries_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\014QueriesProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_QUERIESSERVICE"].methods_by_name["GetQueryPerformanceSummary"]._options = None
    _globals["_QUERIESSERVICE"].methods_by_name["GetQueryPerformanceSummary"]._serialized_options = b"\200}\006"
    _globals["_QUERIESSERVICE"].methods_by_name["ListQueryErrors"]._options = None
    _globals["_QUERIESSERVICE"].methods_by_name["ListQueryErrors"]._serialized_options = b"\200}\006"
    _globals["_QUERIESSERVICE"].methods_by_name["GetQueryErrorsChart"]._options = None
    _globals["_QUERIESSERVICE"].methods_by_name["GetQueryErrorsChart"]._serialized_options = b"\200}\006"
    _globals["_QUERIESSERVICE"].methods_by_name["GetQueryPlan"]._options = None
    _globals["_QUERIESSERVICE"].methods_by_name["GetQueryPlan"]._serialized_options = b"\200}\006"
    _globals["_GETQUERYPERFORMANCESUMMARYREQUEST"]._serialized_start = 191
    _globals["_GETQUERYPERFORMANCESUMMARYREQUEST"]._serialized_end = 261
    _globals["_GETQUERYPERFORMANCESUMMARYRESPONSE"]._serialized_start = 264
    _globals["_GETQUERYPERFORMANCESUMMARYRESPONSE"]._serialized_end = 413
    _globals["_LISTQUERYERRORSPAGETOKEN"]._serialized_start = 416
    _globals["_LISTQUERYERRORSPAGETOKEN"]._serialized_end = 556
    _globals["_QUERYERRORFILTERS"]._serialized_start = 559
    _globals["_QUERYERRORFILTERS"]._serialized_end = 797
    _globals["_QUERYERRORMETA"]._serialized_start = 800
    _globals["_QUERYERRORMETA"]._serialized_end = 1710
    _globals["_LISTQUERYERRORSREQUEST"]._serialized_start = 1713
    _globals["_LISTQUERYERRORSREQUEST"]._serialized_end = 2012
    _globals["_LISTQUERYERRORSRESPONSE"]._serialized_start = 2015
    _globals["_LISTQUERYERRORSRESPONSE"]._serialized_end = 2173
    _globals["_GETQUERYERRORSCHARTREQUEST"]._serialized_start = 2176
    _globals["_GETQUERYERRORSCHARTREQUEST"]._serialized_end = 2535
    _globals["_GETQUERYERRORSCHARTRESPONSE"]._serialized_start = 2537
    _globals["_GETQUERYERRORSCHARTRESPONSE"]._serialized_end = 2626
    _globals["_GETQUERYPLANREQUEST"]._serialized_start = 2628
    _globals["_GETQUERYPLANREQUEST"]._serialized_end = 2685
    _globals["_QUERYPLAN"]._serialized_start = 2688
    _globals["_QUERYPLAN"]._serialized_end = 2881
    _globals["_GETQUERYPLANRESPONSE"]._serialized_start = 2883
    _globals["_GETQUERYPLANRESPONSE"]._serialized_end = 2964
    _globals["_QUERIESSERVICE"]._serialized_start = 2967
    _globals["_QUERIESSERVICE"]._serialized_end = 3448
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import base64
import collections.abc
import ipaddress
import math
import uuid
from datetime import date, datetime, time, timedelta
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union, cast

import cattrs
import dateutil.parser
import isodate
from typing_extensions import get_args, get_origin

from chalk.features._encoding.primitive import ChalkStructType, TPrimitive
from chalk.features._encoding.rich import is_chalk_struct
from chalk.utils.json import TJSON

if TYPE_CHECKING:
    from pydantic import BaseModel
else:
    try:
        from pydantic.v1 import BaseModel
    except ImportError:
        from pydantic import BaseModel

__all__ = ["unstructure_primitive_to_json", "structure_json_to_primitive"]


def unstructure_primitive_to_json(val: TPrimitive) -> TJSON:
    return _default_feature_json_converter.unstructure_primitive_to_json(val)


def structure_json_to_primitive(val: Union[TJSON, TPrimitive], typ: Type[TPrimitive]) -> TPrimitive:
    return _default_feature_json_converter.structure_json_to_primitive(val, typ)


class FeatureJsonConverter:
    def __init__(self, options: FeatureEncodingOptions):
        super().__init__()
        self._json_converter = cattrs.Converter()
        self._options = options
        #######
        # Dicts
        #######

        # All structs get mapped to a typeddict primitive type
        # However, on the wire, we serialize typedicts as lists by field order, similar
        # to how structs are commonly serialized by values
        # Hence, the JSON converter needs to convert all dicts into lists of the values
        # In py3.7+, all dicts are ordered. So, simply iterating over the dict values is sufficient

        if self._options.encode_structs_as_objects:
            self._json_converter.register_unstructure_hook(
                dict, lambda x: {k: self._json_converter.unstructure(v) for k, v in x.items()}
            )
        else:
            self._json_converter.register_unstructure_hook(
                dict, lambda x: [self._json_converter.unstructure(y) for y in x.values()]
            )

        def _structure_dicts(obj: Any, typ: Type):
            # try to treat the incoming 'val' as a dict if typ is a dict
            # ie: if typ == typing.Dict[str, typing.Optional[str]], treat value of [["a","b"],["c","d"]] as {"a": "b", "c": "d"}
            if isinstance(obj, list):
                obj = dict(obj)

            args = get_args(typ)
            if len(args) == 0:
                raise TypeError(
                    f"{typ} types must be parameterized with the key and value types -- for example, `{typ}[str, int]`"
                )
            elif len(args) != 2:
                raise TypeError(f"{typ} should be parameterized with two types, found: {typ}")

            return {k: self._json_converter.structure(v, args[1]) for k, v in obj.items()}

        def _is_dict(typ: Type):
            origin = get_origin(typ)
            return origin in (dict, Dict)

        self._json_converter.register_structure_hook_func(_is_dict, _structure_dicts)

        def _structure_chalk_struct(obj: Any, typ: Type):
            if not isinstance(typ, ChalkStructType):
                raise TypeError(f"Expected ChalkStructType instead of {typ}")
            type_hints = typ.__chalk_type_hints__
            if obj is None:
                return {
                    field_name: self._json_converter.structure(None, type_hint)
                    for (field_name, type_hint) in type_hints.items()
                }
            if isinstance(obj, collections.abc.Mapping):
                # If given a dict, assume it is the primitive type being passed in as the json type
                return {
                    field_name: self._json_converter.structure(obj.get(field_name), type_hint)
                    for (field_name, type_hint) in type_hints.items()
                }
            if not isinstance(obj, collections.abc.Sequence):
                raise TypeError(f"Expected structs to be serialized as lists. Object `{obj}` is not a sequence.")

            if len(type_hints) != len(obj):
                raise TypeError(
                    f"Unable to structure object `{obj}` of size {len(obj)} into type `{typ.__name__}` of size {len(type_hints)}. Size mismatch."
                )

            kwargs = {
                field_name: None if x is None else self._json_converter.structure(x, type_hints[field_name])
                for (x, field_name) in zip(obj, type_hints.keys())
            }
            return kwargs

        self._json_converter.register_structure_hook_func(is_chalk_struct, _structure_chalk_struct)

        #######
        # Lists
        #######

        # The default unstructure hook is fine
        # However, when structuring, we want to allow None for annotations of list[...]

        def _structure_list(obj: Optional[List], typ: Type[List]) -> List:
            args = get_args(typ)

            if len(args) < 1:
                raise TypeError(
                    f"{typ} types must be parameterized with the type of the contained value -- for example, `{typ}[int]`"
                )
            if len(args) > 1:
                raise TypeError(f"{typ} should be parameterized with only one type")
            if obj is None:
                return cast(List, None)
            if not isinstance(obj, (list, tuple)):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError(f"Expected a list, Object `{obj}` is not a list.")

            inner_typ = args[0]
            return [self._json_converter.structure(x, inner_typ) for x in obj]

        def _is_list(typ: Type):
            origin = get_origin(typ)
            return origin in (list, List)

        self._json_converter.register_structure_hook_func(_is_list, _structure_list)

        ######
        # Date
        ######
        def _is_date(x: Type):
            return (
                isinstance(x, type)  # pyright: ignore[reportUnnecessaryIsInstance]
                and issubclass(x, date)
                and not issubclass(x, datetime)  # pyright: ignore[reportUnnecessaryIsInstance]
            )

        self._json_converter.register_unstructure_hook_func(
            _is_date,
            lambda x: x.isoformat(),
        )

        def _structure_date(obj: Any, typ: Type):
            if obj is None:
                return None
            if isinstance(obj, datetime):
                return obj.date()
            if isinstance(obj, date):
                return obj
            if not isinstance(obj, str):
                raise TypeError(
                    f"Date values must be serialized as ISO strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
                )
            return isodate.parse_date(obj)

        self._json_converter.register_structure_hook_func(
            _is_date,
            _structure_date,
        )

        ##########
        # Datetime
        ##########

        self._json_converter.register_unstructure_hook(datetime, lambda x: x.isoformat())

        def _structure_datetime(obj: Any, typ: Type):
            if obj is None:
                return None
            if isinstance(obj, datetime):
                return obj
            if isinstance(obj, date):
                # Upgrade to a datetime
                return datetime.combine(obj, time())
            if not isinstance(obj, str):
                raise TypeError(
                    f"Datetime values must be serialized as ISO strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
                )
            return dateutil.parser.parse(obj)

        self._json_converter.register_structure_hook(datetime, _structure_datetime)

        ######
        # Time
        ######

        self._json_converter.register_unstructure_hook(time, lambda x: x.isoformat())

        def _structure_time(obj: Any, typ: Type):
            if obj is None:
                return None
            if isinstance(obj, time):
                return obj
            if not isinstance(obj, str):
                raise TypeError(
                    f"Time values must be serialized as ISO strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
                )
            return isodate.parse_time(obj)

        self._json_converter.register_structure_hook(time, _structure_time)

        ########
        # Binary
        ########

        self._json_converter.register_unstructure_hook(bytes, lambda x: base64.b64encode(x).decode("utf8"))

        def _structure_bytes(obj: Any, typ: Type):
            if obj is None:
                return None
            if isinstance(obj, str):
                return base64.b64decode(obj)
            if isinstance(obj, bytes):
                return obj
            raise TypeError(
                f"Byte values must be bytes objects or Base64-encoded strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
            )

        self._json_converter.register_structure_hook(bytes, _structure_bytes)

        self._json_converter.register_unstructure_hook(timedelta, isodate.duration_isoformat)

        ###########
        # Timedelta
        ###########

        def _structure_timedelta(obj: Any, typ: Type):
            if obj is None:
                return None
            if isinstance(obj, timedelta):
                return obj
            if not isinstance(obj, str):
                raise TypeError(
                    f"Timedelta values should be serialized as strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
                )
            return isodate.parse_duration(obj)

        self._json_converter.register_structure_hook(timedelta, _structure_timedelta)

        ###########
        # uuid.UUID
        ###########

        def _structure_uuid(obj: Any, typ: Type):
            if obj is None:
                return None
            if isinstance(obj, uuid.UUID):
                return obj
            if not isinstance(obj, str):
                raise TypeError(
                    f"UUID values should be serialized as strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
                )
            return uuid.UUID(obj)

        self._json_converter.register_structure_hook(uuid.UUID, _structure_uuid)

        ###########
        # ipaddress.IPv4Address
        ###########

        def _structure_ipv4(obj: Any, typ: Type):
            if obj is None:
                return None
            if isinstance(obj, ipaddress.IPv4Address):
                return obj
            if isinstance(obj, (int, str)):
                return ipaddress.IPv4Address(obj)
            raise TypeError(f"IPv4Address values should be serialized as strings or integers. Received {obj}")

        self._json_converter.register_structure_hook(ipaddress.IPv4Address, _structure_ipv4)

        ###########
        # ipaddress.IPv6Address
        ###########

        def _structure_ipv6(obj: Any, typ: Type):
            if obj is None:
                return None
            if isinstance(obj, ipaddress.IPv6Address):
                return obj
            if isinstance(obj, (int, str)):
                return ipaddress.IPv6Address(obj)
            raise TypeError(f"IPv6Address values should be serialized as strings or integers. Received {obj}")

        self._json_converter.register_structure_hook(ipaddress.IPv6Address, _structure_ipv6)

        ####################
        # Int/float/str/bool
        ####################

        def _structure_basic(obj: Any, typ: Type):
            if obj is None:
                # Always allow None, even if the field is non-optional
                return None
            if issubclass(typ, bool):
                # For booleans, we are more strict than just doing a simple cast, since
                # bool("random string") or bool(100) should raise an exception, not be True
                if obj in (1, True):
                    return True
                if obj in (0, False):
                    return False
                raise TypeError(f"Cannot convert '{obj}' to a Boolean. Valid values are 1, True, 0, or False.")
            if issubclass(typ, (int, float)):
                if not isinstance(obj, str) and typ(obj) != obj and not math.isnan(obj):
                    raise TypeError(f"Cannot cast '{obj}' of type {type(obj)} to a {typ} without losing precision")
            return typ(obj)  # pyright: ignore[reportCallIssue]

        self._json_converter.register_structure_hook(int, _structure_basic)
        self._json_converter.register_structure_hook(float, _structure_basic)
        self._json_converter.register_structure_hook(str, _structure_basic)
        self._json_converter.register_structure_hook(bool, _structure_basic)

    def unstructure_primitive_to_json(self, val: TPrimitive) -> TJSON:
        return self._json_converter.unstructure(val)

    def structure_json_to_primitive(self, val: Union[TJSON, TPrimitive], typ: Type[TPrimitive]) -> TPrimitive:
        return self._json_converter.structure(val, typ)


######
# Options
######


class FeatureEncodingOptions(BaseModel, frozen=True):
    encode_structs_as_objects: bool = False
    """
    If 'True', a struct type will be encoded as a json object.
    If 'False', a struct type will be encoded as an array where the n-th element
    corresponds to the n-th field of the struct.
    """


_default_feature_json_converter = FeatureJsonConverter(options=FeatureEncodingOptions())
structs_as_objects_feature_json_converter = FeatureJsonConverter(
    options=FeatureEncodingOptions(encode_structs_as_objects=True)
)



from chalk._gen.chalk.arrow.v1 import arrow_pb2 as _arrow_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class TySet(_message.Message):
    __slots__ = ("items",)
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    items: Ty
    def __init__(self, items: _Optional[_Union[Ty, _Mapping]] = ...) -> None: ...

class TyList(_message.Message):
    __slots__ = ("items",)
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    items: Ty
    def __init__(self, items: _Optional[_Union[Ty, _Mapping]] = ...) -> None: ...

class TyIterable(_message.Message):
    __slots__ = ("items",)
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    items: Ty
    def __init__(self, items: _Optional[_Union[Ty, _Mapping]] = ...) -> None: ...

class TyDatetime(_message.Message):
    __slots__ = ("tz",)
    TZ_FIELD_NUMBER: _ClassVar[int]
    tz: str
    def __init__(self, tz: _Optional[str] = ...) -> None: ...

class TyDict(_message.Message):
    __slots__ = ("key", "value")
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: Ty
    value: Ty
    def __init__(
        self, key: _Optional[_Union[Ty, _Mapping]] = ..., value: _Optional[_Union[Ty, _Mapping]] = ...
    ) -> None: ...

class TyTuple(_message.Message):
    __slots__ = ("fixed", "is_variable")
    FIXED_FIELD_NUMBER: _ClassVar[int]
    IS_VARIABLE_FIELD_NUMBER: _ClassVar[int]
    fixed: _containers.RepeatedCompositeFieldContainer[Ty]
    is_variable: bool
    def __init__(self, fixed: _Optional[_Iterable[_Union[Ty, _Mapping]]] = ..., is_variable: bool = ...) -> None: ...

class Ty(_message.Message):
    __slots__ = (
        "nullable",
        "int",
        "str",
        "bool",
        "float",
        "list",
        "set",
        "none",
        "any",
        "iterable",
        "datetime",
        "timedelta",
        "date",
        "tuple",
        "dict",
        "bytes",
    )
    NULLABLE_FIELD_NUMBER: _ClassVar[int]
    INT_FIELD_NUMBER: _ClassVar[int]
    STR_FIELD_NUMBER: _ClassVar[int]
    BOOL_FIELD_NUMBER: _ClassVar[int]
    FLOAT_FIELD_NUMBER: _ClassVar[int]
    LIST_FIELD_NUMBER: _ClassVar[int]
    SET_FIELD_NUMBER: _ClassVar[int]
    NONE_FIELD_NUMBER: _ClassVar[int]
    ANY_FIELD_NUMBER: _ClassVar[int]
    ITERABLE_FIELD_NUMBER: _ClassVar[int]
    DATETIME_FIELD_NUMBER: _ClassVar[int]
    TIMEDELTA_FIELD_NUMBER: _ClassVar[int]
    DATE_FIELD_NUMBER: _ClassVar[int]
    TUPLE_FIELD_NUMBER: _ClassVar[int]
    DICT_FIELD_NUMBER: _ClassVar[int]
    BYTES_FIELD_NUMBER: _ClassVar[int]
    nullable: bool
    int: EmptyMessage
    str: EmptyMessage
    bool: EmptyMessage
    float: EmptyMessage
    list: TyList
    set: TySet
    none: EmptyMessage
    any: EmptyMessage
    iterable: TyIterable
    datetime: TyDatetime
    timedelta: EmptyMessage
    date: EmptyMessage
    tuple: TyTuple
    dict: TyDict
    bytes: EmptyMessage
    def __init__(
        self,
        nullable: bool = ...,
        int: _Optional[_Union[EmptyMessage, _Mapping]] = ...,
        str: _Optional[_Union[EmptyMessage, _Mapping]] = ...,
        bool: _Optional[_Union[EmptyMessage, _Mapping]] = ...,
        float: _Optional[_Union[EmptyMessage, _Mapping]] = ...,
        list: _Optional[_Union[TyList, _Mapping]] = ...,
        set: _Optional[_Union[TySet, _Mapping]] = ...,
        none: _Optional[_Union[EmptyMessage, _Mapping]] = ...,
        any: _Optional[_Union[EmptyMessage, _Mapping]] = ...,
        iterable: _Optional[_Union[TyIterable, _Mapping]] = ...,
        datetime: _Optional[_Union[TyDatetime, _Mapping]] = ...,
        timedelta: _Optional[_Union[EmptyMessage, _Mapping]] = ...,
        date: _Optional[_Union[EmptyMessage, _Mapping]] = ...,
        tuple: _Optional[_Union[TyTuple, _Mapping]] = ...,
        dict: _Optional[_Union[TyDict, _Mapping]] = ...,
        bytes: _Optional[_Union[EmptyMessage, _Mapping]] = ...,
    ) -> None: ...

class EmptyMessage(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class SymbolicConst(_message.Message):
    __slots__ = ("ty", "value")
    TY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    ty: Ty
    value: _arrow_pb2.ScalarValue
    def __init__(
        self,
        ty: _Optional[_Union[Ty, _Mapping]] = ...,
        value: _Optional[_Union[_arrow_pb2.ScalarValue, _Mapping]] = ...,
    ) -> None: ...

class CodeVariable(_message.Message):
    __slots__ = ("name", "module")
    NAME_FIELD_NUMBER: _ClassVar[int]
    MODULE_FIELD_NUMBER: _ClassVar[int]
    name: str
    module: str
    def __init__(self, name: _Optional[str] = ..., module: _Optional[str] = ...) -> None: ...

class CodeVariableValue(_message.Message):
    __slots__ = ("variable", "value")
    VARIABLE_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    variable: CodeVariable
    value: SymbolicConst
    def __init__(
        self,
        variable: _Optional[_Union[CodeVariable, _Mapping]] = ...,
        value: _Optional[_Union[SymbolicConst, _Mapping]] = ...,
    ) -> None: ...

class GlobalVariablesInfo(_message.Message):
    __slots__ = ("code_variables", "environment_variables")
    class EnvironmentVariablesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    CODE_VARIABLES_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_VARIABLES_FIELD_NUMBER: _ClassVar[int]
    code_variables: _containers.RepeatedCompositeFieldContainer[CodeVariableValue]
    environment_variables: _containers.ScalarMap[str, str]
    def __init__(
        self,
        code_variables: _Optional[_Iterable[_Union[CodeVariableValue, _Mapping]]] = ...,
        environment_variables: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/auth/v1/agent.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import featurepermission_pb2 as chalk_dot_auth_dot_v1_dot_featurepermission__pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x19\x63halk/auth/v1/agent.proto\x12\rchalk.auth.v1\x1a%chalk/auth/v1/featurepermission.proto\x1a\x1f\x63halk/auth/v1/permissions.proto"\xee\x01\n\x16\x45nvironmentPermissions\x12;\n\x0bpermissions\x18\x02 \x03(\x0e\x32\x19.chalk.auth.v1.PermissionR\x0bpermissions\x12R\n\x13\x66\x65\x61ture_permissions\x18\x03 \x01(\x0b\x32!.chalk.auth.v1.FeaturePermissionsR\x12\x66\x65\x61turePermissions\x12\x43\n\x0f\x63ustomer_claims\x18\x04 \x03(\x0b\x32\x1a.chalk.auth.v1.CustomClaimR\x0e\x63ustomerClaims"\xec\x02\n\tUserAgent\x12\x1f\n\tclient_id\x18\x01 \x01(\tB\x02\x18\x01R\x08\x63lientId\x12\x17\n\x07user_id\x18\x02 \x01(\tR\x06userId\x12\x17\n\x07team_id\x18\x03 \x01(\tR\x06teamId\x12t\n\x1apermissions_by_environment\x18\x04 \x03(\x0b\x32\x36.chalk.auth.v1.UserAgent.PermissionsByEnvironmentEntryR\x18permissionsByEnvironment\x12"\n\x0cimpersonated\x18\x05 \x01(\x08R\x0cimpersonated\x1ar\n\x1dPermissionsByEnvironmentEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12;\n\x05value\x18\x02 \x01(\x0b\x32%.chalk.auth.v1.EnvironmentPermissionsR\x05value:\x02\x38\x01"7\n\x0b\x43ustomClaim\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x16\n\x06values\x18\x02 \x03(\tR\x06values"\xfa\x02\n\x11ServiceTokenAgent\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x1b\n\tclient_id\x18\x02 \x01(\tR\x08\x63lientId\x12\x17\n\x07team_id\x18\x03 \x01(\tR\x06teamId\x12 \n\x0b\x65nvironment\x18\x04 \x01(\tR\x0b\x65nvironment\x12;\n\x0bpermissions\x18\x05 \x03(\x0e\x32\x19.chalk.auth.v1.PermissionR\x0bpermissions\x12\'\n\rcustom_claims\x18\x06 \x03(\tB\x02\x18\x01R\x0c\x63ustomClaims\x12\x43\n\x0f\x63ustomer_claims\x18\x07 \x03(\x0b\x32\x1a.chalk.auth.v1.CustomClaimR\x0e\x63ustomerClaims\x12R\n\x13\x66\x65\x61ture_permissions\x18\x08 \x01(\x0b\x32!.chalk.auth.v1.FeaturePermissionsR\x12\x66\x65\x61turePermissions"\xa0\x01\n\x0b\x45ngineAgent\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x17\n\x07team_id\x18\x02 \x01(\tR\x06teamId\x12\x1d\n\nproject_id\x18\x03 \x01(\tR\tprojectId\x12%\n\x0e\x65nvironment_id\x18\x04 \x01(\tR\renvironmentId\x12"\n\x0cimpersonated\x18\x05 \x01(\x08R\x0cimpersonated"\x16\n\x14MetadataServiceAgent"\r\n\x0bTenantAgent"\xfe\x02\n\x05\x41gent\x12\x39\n\nuser_agent\x18\x01 \x01(\x0b\x32\x18.chalk.auth.v1.UserAgentH\x00R\tuserAgent\x12R\n\x13service_token_agent\x18\x02 \x01(\x0b\x32 .chalk.auth.v1.ServiceTokenAgentH\x00R\x11serviceTokenAgent\x12?\n\x0c\x65ngine_agent\x18\x03 \x01(\x0b\x32\x1a.chalk.auth.v1.EngineAgentH\x00R\x0b\x65ngineAgent\x12?\n\x0ctenant_agent\x18\x04 \x01(\x0b\x32\x1a.chalk.auth.v1.TenantAgentH\x00R\x0btenantAgent\x12[\n\x16metadata_service_agent\x18\x05 \x01(\x0b\x32#.chalk.auth.v1.MetadataServiceAgentH\x00R\x14metadataServiceAgentB\x07\n\x05\x61gentB\x85\x01\n\x11\x63om.chalk.auth.v1B\nAgentProtoP\x01Z\x0e\x61uth/v1;authv1\xa2\x02\x03\x43\x41X\xaa\x02\rChalk.Auth.V1\xca\x02\rChalk\\Auth\\V1\xe2\x02\x19\x43halk\\Auth\\V1\\GPBMetadata\xea\x02\x0f\x43halk::Auth::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.auth.v1.agent_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\021com.chalk.auth.v1B\nAgentProtoP\001Z\016auth/v1;authv1\242\002\003CAX\252\002\rChalk.Auth.V1\312\002\rChalk\\Auth\\V1\342\002\031Chalk\\Auth\\V1\\GPBMetadata\352\002\017Chalk::Auth::V1"
    _globals["_USERAGENT_PERMISSIONSBYENVIRONMENTENTRY"]._options = None
    _globals["_USERAGENT_PERMISSIONSBYENVIRONMENTENTRY"]._serialized_options = b"8\001"
    _globals["_USERAGENT"].fields_by_name["client_id"]._options = None
    _globals["_USERAGENT"].fields_by_name["client_id"]._serialized_options = b"\030\001"
    _globals["_SERVICETOKENAGENT"].fields_by_name["custom_claims"]._options = None
    _globals["_SERVICETOKENAGENT"].fields_by_name["custom_claims"]._serialized_options = b"\030\001"
    _globals["_ENVIRONMENTPERMISSIONS"]._serialized_start = 117
    _globals["_ENVIRONMENTPERMISSIONS"]._serialized_end = 355
    _globals["_USERAGENT"]._serialized_start = 358
    _globals["_USERAGENT"]._serialized_end = 722
    _globals["_USERAGENT_PERMISSIONSBYENVIRONMENTENTRY"]._serialized_start = 608
    _globals["_USERAGENT_PERMISSIONSBYENVIRONMENTENTRY"]._serialized_end = 722
    _globals["_CUSTOMCLAIM"]._serialized_start = 724
    _globals["_CUSTOMCLAIM"]._serialized_end = 779
    _globals["_SERVICETOKENAGENT"]._serialized_start = 782
    _globals["_SERVICETOKENAGENT"]._serialized_end = 1160
    _globals["_ENGINEAGENT"]._serialized_start = 1163
    _globals["_ENGINEAGENT"]._serialized_end = 1323
    _globals["_METADATASERVICEAGENT"]._serialized_start = 1325
    _globals["_METADATASERVICEAGENT"]._serialized_end = 1347
    _globals["_TENANTAGENT"]._serialized_start = 1349
    _globals["_TENANTAGENT"]._serialized_end = 1362
    _globals["_AGENT"]._serialized_start = 1365
    _globals["_AGENT"]._serialized_end = 1747
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from chalk.integrations.catalogs.base_catalog import BaseCatalog

if TYPE_CHECKING:
    import pyarrow

    from chalk.client import DatasetRevision


def _pyarrow_type_to_hive_type_or_throw(pyarrow_type: pyarrow.DataType) -> str:
    import pyarrow

    if pyarrow_type == pyarrow.utf8():
        return "string"
    elif pyarrow_type == pyarrow.large_utf8():
        return "string"
    elif pyarrow_type == pyarrow.int32():
        return "int"
    elif pyarrow_type == pyarrow.int64():
        return "bigint"
    elif pyarrow_type == pyarrow.uint32():
        return "int"
    elif pyarrow_type == pyarrow.uint64():
        return "bigint"
    elif pyarrow_type == pyarrow.float32():
        return "float"
    elif pyarrow_type == pyarrow.float64():
        return "double"
    elif isinstance(pyarrow_type, pyarrow.TimestampType):
        return "timestamp"
    elif pyarrow_type == pyarrow.date32():
        return "date"
    elif pyarrow_type == pyarrow.bool_():
        return "boolean"
    elif pyarrow_type == pyarrow.Decimal128Type:
        return "decimal"
    elif pyarrow_type == pyarrow.ListType:
        return "array"
    elif pyarrow_type == pyarrow.StructType:
        return "struct"
    elif pyarrow_type == pyarrow.MapType:
        return "map"
    else:
        raise ValueError(f"Unsupported Arrow type '{pyarrow_type}' for conversion to Hive type")


def _convert_arrow_schema_to_athena_schema(arrow_schema: pyarrow.Schema) -> list[dict]:
    athena_schema = []
    for field in list(arrow_schema):
        try:
            type = _pyarrow_type_to_hive_type_or_throw(field.type)
            athena_schema.append(
                {
                    "Name": field.name,
                    "Type": type,
                }
            )
        except ValueError as e:
            raise ValueError(f"Error converting field {field.name} to Athena schema") from e

    return athena_schema


class GlueCatalog(BaseCatalog):
    """
    A class to represent an AWS Glue Catalog.

    Attributes:
        name (str): The name of the Glue Catalog.
        skip_default_database (bool): Whether to skip the default database. Defaults to True.
        aws_role_arn (str | None): The AWS role ARN to assume for accessing the Glue Catalog.
        aws_region (str | None): The AWS region where the Glue Catalog is located.
        aws_profile (str | None): The AWS profile to use for accessing the Glue Catalog.
        aws_secret_key_id (str | None): The AWS secret key ID for accessing the Glue Catalog.
        aws_secret_access_key (str | None): The AWS secret access key for accessing the Glue Catalog.
        aws_session_token (str | None): The AWS session token for accessing the Glue Catalog.
        endpoint_url (str | None): The endpoint URL for the Glue Catalog.
        catalog_id (str | None): The catalog ID for the Glue Catalog.
        cache_ttl (int): The cache time-to-live in seconds. Defaults to 600.

    Methods:
        __init__: Initializes the GlueCatalog with the provided parameters.
    """

    def __init__(
        self,
        name: str,
        skip_default_database: bool = True,
        aws_role_arn: str | None = None,
        aws_region: str | None = None,
        aws_profile: str | None = None,
        aws_secret_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        endpoint_url: str | None = None,
        catalog_id: str | None = None,
        cache_ttl: int = 600,
        glue_client: Any | None = None,
    ) -> None:
        super().__init__()
        self.name = name
        self.cache_ttl = cache_ttl
        self.skip_default_database = skip_default_database
        self.aws_role_arn = aws_role_arn
        self.aws_region = aws_region
        self.aws_profile = aws_profile
        self.aws_secret_key_id = aws_secret_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.endpoint_url = endpoint_url
        self.catalog_id = catalog_id
        self.glue_client = glue_client

    def to_scan_options(self):
        return {
            "catalog_type": "glue",
            "name": self.name,
            "aws_role_arn": self.aws_role_arn,
            "aws_region": self.aws_region,
            "aws_profile": self.aws_profile,
            "endpoint_url": self.endpoint_url,
            "catalog_id": self.catalog_id,
        }

    def write_to_catalog(
        self,
        revision: DatasetRevision,
        destination: str,
    ):
        import boto3

        # Destination must of the form 'database_name.table_name'
        database_name, table_name = destination.split(".")

        glue_client = self.glue_client or boto3.client("glue")

        glue_client.create_table(
            DatabaseName=database_name,
            TableInput={
                "Name": table_name,
                "StorageDescriptor": {
                    "Columns": _convert_arrow_schema_to_athena_schema(revision.arrow_schema()),
                    "Location": revision.output_uris,
                    "InputFormat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat",
                    "OutputFormat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetOutputFormat",
                    "SerdeInfo": {
                        "SerializationLibrary": "org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe"
                    },
                },
                "TableType": "EXTERNAL_TABLE",
            },
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/kubernetes/v1/nodes.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1f\x63halk/kubernetes/v1/nodes.proto\x12\x13\x63halk.kubernetes.v1"\x97\x01\n\x13KubernetesNodeTaint\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x19\n\x05value\x18\x02 \x01(\tH\x00R\x05value\x88\x01\x01\x12\x16\n\x06\x65\x66\x66\x65\x63t\x18\x03 \x01(\tR\x06\x65\x66\x66\x65\x63t\x12"\n\ntime_added\x18\x04 \x01(\x03H\x01R\ttimeAdded\x88\x01\x01\x42\x08\n\x06_valueB\r\n\x0b_time_added"\x93\x02\n\x12KubernetesNodeSpec\x12\x1e\n\x08pod_cidr\x18\x01 \x01(\tH\x00R\x07podCidr\x88\x01\x01\x12@\n\x06taints\x18\x02 \x03(\x0b\x32(.chalk.kubernetes.v1.KubernetesNodeTaintR\x06taints\x12\x1b\n\tpod_cidrs\x18\x03 \x03(\tR\x08podCidrs\x12$\n\x0bprovider_id\x18\x04 \x01(\tH\x01R\nproviderId\x88\x01\x01\x12)\n\runschedulable\x18\x05 \x01(\x08H\x02R\runschedulable\x88\x01\x01\x42\x0b\n\t_pod_cidrB\x0e\n\x0c_provider_idB\x10\n\x0e_unschedulable"\xfa\x01\n\x17KubernetesNodeCondition\x12\x12\n\x04type\x18\x01 \x01(\tR\x04type\x12\x16\n\x06status\x18\x02 \x01(\tR\x06status\x12.\n\x13last_heartbeat_time\x18\x03 \x01(\x03R\x11lastHeartbeatTime\x12\x30\n\x14last_transition_time\x18\x04 \x01(\x03R\x12lastTransitionTime\x12\x1b\n\x06reason\x18\x05 \x01(\tH\x00R\x06reason\x88\x01\x01\x12\x1d\n\x07message\x18\x06 \x01(\tH\x01R\x07message\x88\x01\x01\x42\t\n\x07_reasonB\n\n\x08_message"O\n\x18KubernetesAttachedVolume\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x1f\n\x0b\x64\x65vice_path\x18\x02 \x01(\tR\ndevicePath"\xe4\x01\n\x14KubernetesNodeStatus\x12L\n\nconditions\x18\x01 \x03(\x0b\x32,.chalk.kubernetes.v1.KubernetesNodeConditionR\nconditions\x12$\n\x0evolumes_in_use\x18\x02 \x03(\tR\x0cvolumesInUse\x12X\n\x10volumes_attached\x18\x03 \x03(\x0b\x32-.chalk.kubernetes.v1.KubernetesAttachedVolumeR\x0fvolumesAttached"\xde\x08\n\x12KubernetesNodeData\x12\x12\n\x04team\x18\x01 \x01(\tR\x04team\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12\x10\n\x03uid\x18\x03 \x01(\tR\x03uid\x12#\n\rinstance_type\x18\x04 \x01(\tR\x0cinstanceType\x12\x16\n\x06region\x18\x05 \x01(\tR\x06region\x12\x12\n\x04zone\x18\x06 \x01(\tR\x04zone\x12-\n\x12\x63reation_timestamp\x18\x08 \x01(\x03R\x11\x63reationTimestamp\x12-\n\x12\x64\x65letion_timestamp\x18\t \x01(\x03R\x11\x64\x65letionTimestamp\x12-\n\x12observed_timestamp\x18\n \x01(\x03R\x11observedTimestamp\x12K\n\x06labels\x18\x0b \x03(\x0b\x32\x33.chalk.kubernetes.v1.KubernetesNodeData.LabelsEntryR\x06labels\x12Z\n\x0b\x61nnotations\x18\x0c \x03(\x0b\x32\x38.chalk.kubernetes.v1.KubernetesNodeData.AnnotationsEntryR\x0b\x61nnotations\x12\x1d\n\nmachine_id\x18\r \x01(\tR\tmachineId\x12\x1f\n\x0bsystem_uuid\x18\x0e \x01(\tR\nsystemUuid\x12\x17\n\x07\x62oot_id\x18\x0f \x01(\tR\x06\x62ootId\x12$\n\runschedulable\x18\x10 \x01(\x08R\runschedulable\x12\x1c\n\tnamespace\x18\x11 \x01(\tR\tnamespace\x12\x1f\n\x0binstance_id\x18\x12 \x01(\tR\ninstanceId\x12\x18\n\x07\x63luster\x18\x13 \x01(\tR\x07\x63luster\x12\x1b\n\ttotal_cpu\x18\x14 \x01(\tR\x08totalCpu\x12!\n\x0ctotal_memory\x18\x15 \x01(\tR\x0btotalMemory\x12\'\n\x0f\x61llocatable_cpu\x18\x16 \x01(\tR\x0e\x61llocatableCpu\x12-\n\x12\x61llocatable_memory\x18\x17 \x01(\tR\x11\x61llocatableMemory\x12@\n\x04spec\x18\x18 \x01(\x0b\x32\'.chalk.kubernetes.v1.KubernetesNodeSpecH\x00R\x04spec\x88\x01\x01\x12\x46\n\x06status\x18\x19 \x01(\x0b\x32).chalk.kubernetes.v1.KubernetesNodeStatusH\x01R\x06status\x88\x01\x01\x1a\x39\n\x0bLabelsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a>\n\x10\x41nnotationsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x07\n\x05_specB\t\n\x07_statusB\xaf\x01\n\x17\x63om.chalk.kubernetes.v1B\nNodesProtoP\x01Z\x1akubernetes/v1;kubernetesv1\xa2\x02\x03\x43KX\xaa\x02\x13\x43halk.Kubernetes.V1\xca\x02\x13\x43halk\\Kubernetes\\V1\xe2\x02\x1f\x43halk\\Kubernetes\\V1\\GPBMetadata\xea\x02\x15\x43halk::Kubernetes::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.kubernetes.v1.nodes_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\027com.chalk.kubernetes.v1B\nNodesProtoP\001Z\032kubernetes/v1;kubernetesv1\242\002\003CKX\252\002\023Chalk.Kubernetes.V1\312\002\023Chalk\\Kubernetes\\V1\342\002\037Chalk\\Kubernetes\\V1\\GPBMetadata\352\002\025Chalk::Kubernetes::V1"
    _globals["_KUBERNETESNODEDATA_LABELSENTRY"]._options = None
    _globals["_KUBERNETESNODEDATA_LABELSENTRY"]._serialized_options = b"8\001"
    _globals["_KUBERNETESNODEDATA_ANNOTATIONSENTRY"]._options = None
    _globals["_KUBERNETESNODEDATA_ANNOTATIONSENTRY"]._serialized_options = b"8\001"
    _globals["_KUBERNETESNODETAINT"]._serialized_start = 57
    _globals["_KUBERNETESNODETAINT"]._serialized_end = 208
    _globals["_KUBERNETESNODESPEC"]._serialized_start = 211
    _globals["_KUBERNETESNODESPEC"]._serialized_end = 486
    _globals["_KUBERNETESNODECONDITION"]._serialized_start = 489
    _globals["_KUBERNETESNODECONDITION"]._serialized_end = 739
    _globals["_KUBERNETESATTACHEDVOLUME"]._serialized_start = 741
    _globals["_KUBERNETESATTACHEDVOLUME"]._serialized_end = 820
    _globals["_KUBERNETESNODESTATUS"]._serialized_start = 823
    _globals["_KUBERNETESNODESTATUS"]._serialized_end = 1051
    _globals["_KUBERNETESNODEDATA"]._serialized_start = 1054
    _globals["_KUBERNETESNODEDATA"]._serialized_end = 2172
    _globals["_KUBERNETESNODEDATA_LABELSENTRY"]._serialized_start = 2031
    _globals["_KUBERNETESNODEDATA_LABELSENTRY"]._serialized_end = 2088
    _globals["_KUBERNETESNODEDATA_ANNOTATIONSENTRY"]._serialized_start = 2090
    _globals["_KUBERNETESNODEDATA_ANNOTATIONSENTRY"]._serialized_end = 2152
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/cli.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x19\x63halk/server/v1/cli.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto"\x9e\x01\n\x1b\x43ommandLineInterfaceVersion\x12\x18\n\x07version\x18\x01 \x01(\tR\x07version\x12!\n\x0c\x64ownload_url\x18\x02 \x01(\tR\x0b\x64ownloadUrl\x12\x0e\n\x02os\x18\x03 \x01(\tR\x02os\x12\x12\n\x04\x61rch\x18\x04 \x01(\tR\x04\x61rch\x12\x1e\n\ngeneration\x18\x05 \x01(\x03R\ngeneration"R\n\x12GetVersionsRequest\x12\x13\n\x02os\x18\x01 \x01(\tH\x00R\x02os\x88\x01\x01\x12\x17\n\x04\x61rch\x18\x02 \x01(\tH\x01R\x04\x61rch\x88\x01\x01\x42\x05\n\x03_osB\x07\n\x05_arch"\x87\x02\n\x13GetVersionsResponse\x12H\n\x08versions\x18\x01 \x03(\x0b\x32,.chalk.server.v1.CommandLineInterfaceVersionR\x08versions\x12\x44\n\x06latest\x18\x02 \x01(\x0b\x32,.chalk.server.v1.CommandLineInterfaceVersionR\x06latest\x12\x46\n\x07nightly\x18\x03 \x01(\x0b\x32,.chalk.server.v1.CommandLineInterfaceVersionR\x07nightly\x12\x18\n\x07minimum\x18\x04 \x01(\tR\x07minimum2\x7f\n\x1b\x43ommandLineInterfaceService\x12`\n\x0bGetVersions\x12#.chalk.server.v1.GetVersionsRequest\x1a$.chalk.server.v1.GetVersionsResponse"\x06\x90\x02\x01\x80}\x01\x42\x91\x01\n\x13\x63om.chalk.server.v1B\x08\x43liProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.cli_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\010CliProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_COMMANDLINEINTERFACESERVICE"].methods_by_name["GetVersions"]._options = None
    _globals["_COMMANDLINEINTERFACESERVICE"].methods_by_name[
        "GetVersions"
    ]._serialized_options = b"\220\002\001\200}\001"
    _globals["_COMMANDLINEINTERFACEVERSION"]._serialized_start = 80
    _globals["_COMMANDLINEINTERFACEVERSION"]._serialized_end = 238
    _globals["_GETVERSIONSREQUEST"]._serialized_start = 240
    _globals["_GETVERSIONSREQUEST"]._serialized_end = 322
    _globals["_GETVERSIONSRESPONSE"]._serialized_start = 325
    _globals["_GETVERSIONSRESPONSE"]._serialized_end = 588
    _globals["_COMMANDLINEINTERFACESERVICE"]._serialized_start = 590
    _globals["_COMMANDLINEINTERFACESERVICE"]._serialized_end = 717
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import uuid
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Collection, Iterable, Literal, Mapping, Optional, Sequence, Union

import requests
from typing_extensions import TypeAlias

from chalk.client.models import (
    BranchDeployResponse,
    BranchIdParam,
    BulkOnlineQueryResponse,
    ChalkError,
    FeatureDropResponse,
    FeatureObservationDeletionResponse,
    FeatureReference,
    FeatureStatisticsResponse,
    GetIncrementalProgressResponse,
    OfflineQueryInputUri,
    OnlineQuery,
    OnlineQueryContext,
    PlanQueryResponse,
    ResolverRunResponse,
    ResourceRequests,
    StreamResolverTestResponse,
    WhoAmIResponse,
)
from chalk.client.response import Dataset, OnlineQueryResult
from chalk.features import DataFrame, Feature

if TYPE_CHECKING:
    import ssl

    import pandas as pd
    import polars as pl
    from pydantic import BaseModel

    QueryInput = Mapping[FeatureReference, Any] | pd.DataFrame | pl.DataFrame | DataFrame

from chalk.features._encoding.json import FeatureEncodingOptions
from chalk.features.resolver import Resolver
from chalk.features.tag import BranchId, DeploymentId, EnvironmentId
from chalk.parsed.branch_state import BranchGraphSummary
from chalk.prompts import Prompt


class ChalkClient:
    """The `ChalkClient` is the primary Python interface for interacting with Chalk.

    You can use it to query data, trigger resolver runs, gather offline data, and more.
    """

    def __init__(
        self,
        client_id: str | None = None,
        client_secret: str | None = None,
        environment: EnvironmentId | None = None,
        api_server: str | None = None,
        branch: BranchId | None | Literal[True] = None,
        deployment_tag: str | None = None,
        preview_deployment_id: DeploymentId | None = None,
        session: requests.Session | None = None,
        query_server: str | None = None,
        additional_headers: Mapping[str, str] | None = None,
        default_job_timeout: float | timedelta | None = None,
        default_request_timeout: float | timedelta | None = None,
        default_connect_timeout: float | timedelta | None = None,
        local: bool = False,
        ssl_context: ssl.SSLContext | None = None,
    ):
        """Create a `ChalkClient` with the given credentials.

        Parameters
        ----------
        client_id
            The client ID to use to authenticate. Can either be a
            service token id or a user token id.
        client_secret
            The client secret to use to authenticate. Can either be a
            service token secret or a user token secret.
        environment
            The ID or name of the environment to use for this client.
            Not necessary if your `client_id` and `client_secret`
            are for a service token scoped to a single environment.
            If not present, the client will use the environment variable
            `CHALK_ENVIRONMENT`.
        api_server
            The API server to use for this client. Required if you are
            using a Chalk Dedicated deployment. If not present, the client
            will check for the presence of the environment variable
            `CHALK_API_SERVER`, and use that if found.
        query_server
            The query server to use for this client. Required if you are
            using a standalone Chalk query engine deployment. If not present,
            the client will default to the value of api_server.
        branch
            If specified, Chalk will route all requests from this client
            instance to the relevant branch. Some methods allow you to
            override this instance-level branch configuration by passing
            in a `branch` argument.

            If `True`, the client will pick up the branch from the
            current git branch.
        deployment_tag
            If specified, Chalk will route all requests from this client
            instance to the relevant tagged deployment. This cannot be
            used with the `branch` argument.
        preview_deployment_id
            If specified, Chalk will route all requests from this client
            instance to the relevant preview deployment.
        session
            A `requests.Session` to use for all requests. If not provided,
            a new session will be created.
        additional_headers
            A map of additional HTTP headers to pass with each request.
        default_job_timeout:
            The default wait timeout, in seconds, to wait for long-running jobs to complete
            when accessing query results.
            Jobs will not time out if this timeout elapses. For no timeout, set to None.
            The default is no timeout.
        default_request_timeout:
            The default wait timeout, in seconds, to wait for network requests to complete.
            If not specified, the default is no timeout.
        default_connect_timeout:
            The default connection timeout, in seconds, to wait for establishing a connection.
            This is separate from the request timeout and controls only the connection phase.
            If not specified, the default is no timeout.
        local
            If True, point the client at a local version of the code.
        ssl_context
            A `ssl.SSLContext` that can be loaded with self-signed certificates so that
            `requests` requests to servers hosted with self-signed certificates succeed.

        Raises
        ------
        ChalkAuthException
            If `client_id` or `client_secret` are not provided, there
            is no `~/.chalk.yml` file with applicable credentials,
            and the environment variables `CHALK_CLIENT_ID` and
            `CHALK_CLIENT_SECRET` are not set.
        """
        super().__init__()
        ...

    def query(
        self,
        input: Mapping[FeatureReference, Any] | Any,
        output: Sequence[FeatureReference] = (),
        now: datetime | None = None,
        staleness: Mapping[FeatureReference, str] | None = None,
        environment: EnvironmentId | None = None,
        tags: list[str] | None = None,
        preview_deployment_id: str | None = None,
        branch: BranchId | None = ...,
        correlation_id: str | None = None,
        query_name: str | None = None,
        query_name_version: str | None = None,
        include_meta: bool = False,
        meta: Mapping[str, str] | None = None,
        explain: bool = False,
        store_plan_stages: bool = False,
        encoding_options: FeatureEncodingOptions | None = None,
        required_resolver_tags: list[str] | None = None,
        planner_options: Mapping[str, Union[str, int, bool]] | None = None,
        request_timeout: Optional[float] = None,
        connect_timeout: Optional[float] = None,
        headers: Mapping[str, str] | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
    ) -> OnlineQueryResult:
        """Compute features values using online resolvers.
        See https://docs.chalk.ai/docs/query-basics for more information.

        Parameters
        ----------
        input
            The features for which there are known values, mapped to those values.
            For example, `{User.id: 1234}`. Features can also be expressed as snakecased strings,
            e.g. `{"user.id": 1234}`
        output
            Outputs are the features that you'd like to compute from the inputs.
            For example, `[User.age, User.name, User.email]`.

            If an empty sequence, the output will be set to all features on the namespace
            of the query. For example, if you pass as input `{"user.id": 1234}`, then the query
            is defined on the `User` namespace, and all features on the `User` namespace
            (excluding has-one and has-many relationships) will be used as outputs.
        staleness
            Maximum staleness overrides for any output features or intermediate features.
            See https://docs.chalk.ai/docs/query-caching for more information.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        tags
            The tags used to scope the resolvers.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        required_resolver_tags
            If specified, *all* required_resolver_tags must be present on a resolver for it to be
            considered eligible to execute.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        branch
            If specified, Chalk will route your request to the relevant branch.
        preview_deployment_id
            If specified, Chalk will route your request to the relevant preview deployment.
        query_name
            The semantic name for the query you're making, for example, `"loan_application_model"`.
            Typically, each query that you make from your application should have a name.
            Chalk will present metrics and dashboard functionality grouped by 'query_name'.
            If your query name matches a `NamedQuery`, the query will automatically pull outputs
            and options specified in the matching `NamedQuery`.
        query_name_version
            If `query_name` is specified, this specifies the version of the named query you're making.
            This is only useful if you want your query to use a `NamedQuery` with a specific name and a
            specific version. If a `query_name` has not been supplied, then this parameter is ignored.
        include_meta
            Returns metadata about the query execution under `OnlineQueryResult.meta`.
            This could make the query slightly slower.
            For more information, see https://docs.chalk.ai/docs/query-basics.
        explain
            Log the query execution plan. Requests using `explain=True` will be slower
            than requests using `explain=False`.

            If `True`, 'include_meta' will be set to `True` as well.
        store_plan_stages
            If `True`, the output of each of the query plan stages will be stored.
            This option dramatically impacts the performance of the query,
            so it should only be used for debugging.
        correlation_id
            You can specify a correlation ID to be used in logs and web interfaces.
            This should be globally unique, i.e. a `uuid` or similar. Logs generated
            during the execution of your query will be tagged with this correlation id.
        now
            The time at which to evaluate the query. If not specified, the current time will be used.
            This parameter is complex in the context of online_query since the online store
            only stores the most recent value of an entity's features. If `now` is in the past,
            it is extremely likely that `None` will be returned for cache-only features.

            This parameter is primarily provided to support:
                - controlling the time window for aggregations over cached has-many relationships
                - controlling the time window for aggregations over has-many relationships loaded from an
                  external database

            If you are trying to perform an exploratory analysis of past feature values, prefer `offline_query`.
        query_context
            An immutable context that can be accessed from Python resolvers.
            This context wraps a JSON-compatible dictionary or JSON string with type restrictions.
            See https://docs.chalk.ai/api-docs#ChalkContext for more information.

        Other Parameters
        ----------------
        meta
            Arbitrary `key:value` pairs to associate with a query.
        planner_options
            Dictionary of additional options to pass to the Chalk query engine.
            Values may be provided as part of conversations with Chalk support
            to enable or disable specific functionality.
        request_timeout
            Float value indicating number of seconds that the request should wait before timing out
            at the network level. May not cancel resources on the server processing the query.
        connect_timeout
            Float value indicating number of seconds to wait for establishing a connection.
            This is separate from request_timeout and controls only the connection phase.
        headers
            Additional headers to provide with the request

        Returns
        -------
        OnlineQueryResult
            Wrapper around the output features and any query metadata
            and errors encountered while running the resolvers.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> result = ChalkClient().query(
        ...     input={
        ...         User.name: "Katherine Johnson"
        ...     },
        ...     output=[User.fico_score],
        ...     staleness={User.fico_score: "10m"},
        ... )
        >>> result.get_feature_value(User.fico_score)
        """
        ...

    def multi_query(
        self,
        queries: list[OnlineQuery],
        environment: EnvironmentId | None = None,
        preview_deployment_id: str | None = None,
        branch: BranchId | None = ...,
        correlation_id: str | None = None,
        query_name: str | None = None,
        query_name_version: str | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
        meta: Mapping[str, str] | None = None,
        use_feather: bool | None = True,
        compression: str | None = "uncompressed",
    ) -> BulkOnlineQueryResponse:
        """
        Execute multiple queries (represented by `queries=` argument) in a single request. This is useful if the
        queries are "rooted" in different `@features` classes -- i.e. if you want to load features for `User` and
        `Merchant` and there is no natural relationship object which is related to both of these classes, `multi_query`
        allows you to submit two independent queries.

        Returns a `BulkOnlineQueryResponse`, which is functionally a list of query results. Each of these result
        can be accessed by index. Individual results can be further checked for errors and converted
        to pandas or polars DataFrames.

        In contrast, `query_bulk` executes a single query with multiple inputs/outputs.

        Parameters
        ----------
        queries
            A list of the `OnlineQuery` objects you'd like to execute.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        branch
            If specified, Chalk will route your request to the relevant branch.
        preview_deployment_id
            If specified, Chalk will route your request to the
            relevant preview deployment.

        Other Parameters
        ----------------
        query_name
            The name for class of query you're making, for example, `"loan_application_model"`.
        query_context
            An immutable context that can be accessed from Python resolvers.
            This context wraps a JSON-compatible dictionary or JSON string with type restrictions.
            See https://docs.chalk.ai/api-docs#ChalkContext for more information.
        correlation_id
            A globally unique ID for the query, used alongside logs and
            available in web interfaces.
        meta
            Arbitrary `key:value` pairs to associate with a query.
        compression
            Which compression scheme to use pyarrow. Options are: `"zstd"`, `"lz4"`, `"uncompressed"`.

        Returns
        -------
        BulkOnlineQueryResponse
            An output containing results as a `list[BulkOnlineQueryResult]`,
            where each result contains a `DataFrame` of the results of each
            query or any errors.

        Examples
        --------
        >>> from chalk.client import ChalkClient, OnlineQuery
        >>> queries = [
        ...     OnlineQuery(
        ...         input={User.name: ['Katherine Johnson']},
        ...         output=[User.fico_score],
        ...     ),
        ...     OnlineQuery(
        ...         input={Merchant.name: ['Eight Sleep']},
        ...         output=[Merchant.address],
        ...     ),
        ... ]
        >>> result = ChalkClient().multi_query(queries)
        >>> result[0].get_feature_value(User.fico_score)
        """
        ...

    def query_bulk(
        self,
        input: Mapping[FeatureReference, Sequence[Any]],
        output: Sequence[FeatureReference] = (),
        now: Sequence[datetime] | None = None,
        staleness: Mapping[FeatureReference, str] | None = None,
        context: OnlineQueryContext | None = None,  # Deprecated.
        environment: EnvironmentId | None = None,
        store_plan_stages: bool = False,
        tags: list[str] | None = None,
        required_resolver_tags: list[str] | None = None,
        preview_deployment_id: str | None = None,
        branch: BranchId | None = ...,
        correlation_id: str | None = None,
        query_name: str | None = None,
        query_name_version: str | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
        meta: Mapping[str, str] | None = None,
        explain: bool = False,
        request_timeout: Optional[float] = None,
        headers: Mapping[str, str] | None = None,
    ) -> BulkOnlineQueryResponse:
        """Compute features values for many rows of inputs using online resolvers.
        See https://docs.chalk.ai/docs/query-basics for more information on online query.

        This method is similar to `query`, except it takes in `list` of inputs, and produces one
        output per row of inputs.

        This method is appropriate if you want to fetch the same set of features for many different
        input primary keys.

        This method contrasts with `multi_query`, which executes multiple fully independent queries.

        This endpoint is not available in all environments.

        Parameters
        ----------
        input
            The features for which there are known values, mapped to a list
            of the values.
        output
            Outputs are the features that you'd like to compute from the inputs.
        staleness
            Maximum staleness overrides for any output features or intermediate features.
            See https://docs.chalk.ai/docs/query-caching for more information.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        tags
            The tags used to scope the resolvers.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        branch
            If specified, Chalk will route your request to the relevant branch.
        preview_deployment_id
            If specified, Chalk will route your request to the
            relevant preview deployment.
        now
            The time at which to evaluate the query. If not specified, the current time will be used.
            The length of this list must be the same as the length of the values in `input`.

        Other Parameters
        ----------------
        query_name
            The semantic name for the query you're making, for example, `"loan_application_model"`.
            Typically, each query that you make from your application should have a name.
            Chalk will present metrics and dashboard functionality grouped by 'query_name'.
            If your query name matches a `NamedQuery`, the query will automatically pull outputs
            and options specified in the matching `NamedQuery`.
        query_name_version
            If `query_name` is specified, this specifies the version of the named query you're making.
            This is only useful if you want your query to use a `NamedQuery` with a specific name and a
            specific version. If a `query_name` has not been supplied, then this parameter is ignored.
        query_context
            An immutable context that can be accessed from Python resolvers.
            This context wraps a JSON-compatible dictionary or JSON string with type restrictions.
            See https://docs.chalk.ai/api-docs#ChalkContext for more information.
        correlation_id
            A globally unique ID for the query, used alongside logs and
            available in web interfaces.
        meta
            Arbitrary `key:value` pairs to associate with a query.
        context
            Deprecated in favor of `environment` and `tags`.
        request_timeout
            Float value indicating number of seconds that the request should wait before timing out
            at the network level. May not cancel resources on the server processing the query
        explain
            Log the query execution plan. Requests using `explain=True` will be slower
            than requests using `explain=False`.
        headers
            Additional headers to provide with the request

        Returns
        -------
        BulkOnlineQueryResponse
            An output containing results as a `list[BulkOnlineQueryResult]`,
            where each result contains a `DataFrame` of the results of each query.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        ... ChalkClient().query_bulk(
        ...     input={User.name: ["Katherine Johnson", "Eleanor Roosevelt"]},
        ...     output=[User.fico_score],
        ...     staleness={User.fico_score: "10m"},
        ... )
        """
        ...

    def plan_query(
        self,
        input: Sequence[FeatureReference],
        output: Sequence[FeatureReference],
        staleness: Mapping[FeatureReference, str] | None = None,
        environment: EnvironmentId | None = None,
        tags: list[str] | None = None,
        preview_deployment_id: str | None = None,
        branch: Union[BranchId, None] = ...,
        query_name: str | None = None,
        query_name_version: str | None = None,
        meta: Mapping[str, str] | None = None,
        store_plan_stages: bool = False,
        explain: bool = False,
        num_input_rows: Optional[int] = None,
        headers: Mapping[str, str] | None = None,
    ) -> PlanQueryResponse:
        """Plan a query without executing it.

        Parameters
        ----------
        input
            The features for which there are known values, mapped to those values.
            For example, `{User.id: 1234}`. Features can also be expressed as snakecased strings,
            e.g. `{"user.id": 1234}`
        output
            Outputs are the features that you'd like to compute from the inputs.
            For example, `[User.age, User.name, User.email]`.
        staleness
            Maximum staleness overrides for any output features or intermediate features.
            See https://docs.chalk.ai/docs/query-caching for more information.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        tags
            The tags used to scope the resolvers.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        branch
            If specified, Chalk will route your request to the relevant branch.
        preview_deployment_id
            If specified, Chalk will route your request to the relevant preview deployment.
        query_name
            The semantic name for the query you're making, for example, `"loan_application_model"`.
            Typically, each query that you make from your application should have a name.
            Chalk will present metrics and dashboard functionality grouped by 'query_name'.
            If your query name matches a `NamedQuery`, the query will automatically pull outputs
            and options specified in the matching `NamedQuery`.
        query_name_version
            If `query_name` is specified, this specifies the version of the named query you're making.
            This is only useful if you want your query to use a `NamedQuery` with a specific name and a
            specific version. If a `query_name` has not been supplied, then this parameter is ignored.
        meta
            Arbitrary `key:value` pairs to associate with a query.
        store_plan_stages
            If true, the plan will store the intermediate values at each stage in the plan
        explain
            If true, the plan will emit additional output to assist with debugging.
        num_input_rows:
            The number of input rows that this plan will be run with. If unknown, specify `None`.
        headers
            Additional headers to provide with the request

        Returns
        -------
        PlanQueryResponse
            The query plan, including the resolver execution order and the
            resolver execution plan for each resolver.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> result = ChalkClient().plan_query(
        ...     input=[User.id],
        ...     output=[User.fico_score],
        ...     staleness={User.fico_score: "10m"},
        ... )
        >>> result.rendered_plan
        >>> result.output_schema
        """
        ...

    def check(
        self,
        input: Mapping[FeatureReference, Any] | Any,
        assertions: Mapping[FeatureReference, Any],
        cache_hits: Iterable[str | Any] | None = None,
        feature_errors: Mapping[str | Any, Any] | None = None,
        query_errors: Optional[Collection[ChalkError]] = None,
        now: datetime | None = None,
        staleness: Mapping[FeatureReference, str] | None = None,
        tags: list[str] | None = None,
        query_name: str | None = None,
        query_name_version: str | None = None,
        encoding_options: FeatureEncodingOptions | None = None,
        required_resolver_tags: list[str] | None = None,
        planner_options: Mapping[str, Union[str, int, bool]] | None = None,
        request_timeout: Optional[float] = None,
        headers: Mapping[str, str] | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
        value_metrics_tag_by_features: Sequence[FeatureReference] = (),
        show_table: bool = False,
        float_rel_tolerance: float = 1e-6,
        float_abs_tolerance: float = 1e-12,
        prefix: bool = True,
        show_matches: bool = True,
    ):
        """Check whether expected results of a query match Chalk query ouputs.
        This function should be used in integration tests.
        If you're using `pytest`, `pytest.fail` will be executed on an error.
        Otherwise, an `AssertionError` will be raised.

        Parameters
        ----------
        input
            A feature set or a mapping of `{feature: value}` of givens.
            All values will be encoded to the json representation.
        assertions
            A feature set or a mapping of `{feature: value}` of expected outputs.
            For values where you do not care about the result, use an `...` for the
            feature value (i.e. when an error is expected).
        cache_hits
            A list of the features that you expect to be read from the online
            store, e.g.
            >>> cache_hits=[Actor.name, Actor.num_appearances]
        feature_errors
            A map from the expected feature name to the expected errors for that feature, e.g.
            >>> expected_feature_errors={
            ...     User.id: [ChalkError(...), ChalkError(...)]
            ... }
            >>> errors={
            ...     "user.id": [ChalkError(...), ChalkError(...)]
            ... }
        query_errors
            A list of the expected query error.
        now
            The time at which to evaluate the query. If not specified, the current time will be used.
            This parameter is complex in the context of `online_query` since the online store
            only stores the most recent value of an entity's features. If `now` is in the past,
            it is extremely likely that `None` will be returned for cache-only features.

            This parameter is primarily provided to support:
                - controlling the time window for aggregations over cached has-many relationships
                - controlling the time window for aggregations over has-many relationships loaded from an
                  external database

            If you are trying to perform an exploratory analysis of past feature values, prefer `offline_query`.
        staleness
            Maximum staleness overrides for any output features or intermediate features.
            See https://docs.chalk.ai/docs/query-caching for more information.
        tags
            The tags used to scope the resolvers.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        query_name
            The semantic name for the query you're making, for example, `"loan_application_model"`.
            Typically, each query that you make from your application should have a name.
            Chalk will present metrics and dashboard functionality grouped by 'query_name'.
            If your query name matches a `NamedQuery`, the query will automatically pull outputs
            and options specified in the matching `NamedQuery`.
        query_name_version
            If `query_name` is specified, this specifies the version of the named query you're making.
            This is only useful if you want your query to use a `NamedQuery` with a specific name and a
            specific version. If a `query_name` has not been supplied, then this parameter is ignored.
        query_context
            An immutable context that can be accessed from Python resolvers.
            This context wraps a JSON-compatible dictionary or JSON string with type restrictions.
            See https://docs.chalk.ai/api-docs#ChalkContext for more information.
        required_resolver_tags
            If specified, *all* required_resolver_tags must be present on a resolver for it to be
            considered eligible to execute.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        show_table
            Print the feature value table even if no errors were found.
        show_matches
            If `True`, show the expected and actual values that match.
            If `False`, only show the expected and actual values that do not match.
        float_rel_tolerance
            The relative tolerenance to allow for float equality.
            If you specify both `float_rel_tolerance` and `float_abs_tolerance`,
            the numbers will be considered equal if either tolerance is met.
            Equivalent to:
            >>> abs(a - b) <= float_rel_tolerance * max(abs(a), abs(b))
        float_abs_tolerance
            The absolute tolerenance to allow for float equality.
            If you specify both `float_rel_tolerance` and `float_abs_tolerance`,
            the numbers will be considered equal if either tolerance is met.
            Equivalent to:
            >>> abs(a - b) <= float_abs_tolerance
        prefix
            Whether to show the prefix for feature names in the table.

        Other Parameters
        ----------------
        planner_options
            Dictionary of additional options to pass to the Chalk query engine.
            Values may be provided as part of conversations with Chalk support
            to enable or disable specific functionality.
        request_timeout
            Float value indicating number of seconds that the request should wait before timing out
            at the network level. May not cancel resources on the server processing the query.
        headers
            Additional headers to provide with the request

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> result = ChalkClient().check(
        ...     input={Actor.id: "nm0000001"},
        ...     assertions={Actor.num_movies: 40},
        ... )
        Chalk Feature Value Mismatch
        ┏━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━┓
        ┃ Kind   ┃ Name                 ┃ Value     ┃
        ┡━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━┩
        │ Expect │ actor.id             │ nm0000001 │
        │ Actual │ actor.id             │ nm0000001 │
        │ Expect │ actor.num_appearanc… │ 40        │
        │ Actual │ actor.num_appearanc… │ 41        │
        └────────┴──────────────────────┴───────────┘
        """
        ...

    def get_dataset(
        self,
        dataset_name: Optional[str] = None,
        environment: Optional[EnvironmentId] = None,
        *,
        dataset_id: str | uuid.UUID | None = None,
        revision_id: str | uuid.UUID | None = None,
        job_id: str | uuid.UUID | None = None,
    ) -> Dataset:
        """Get a Chalk `Dataset` containing data from a previously created dataset.

        If an offline query has been created with a dataset name, `.get_dataset` will
        return a Chalk `Dataset`.
        The `Dataset` wraps a lazily-loading Chalk `DataFrame` that enables us to analyze
        our data without loading all of it directly into memory.
        See https://docs.chalk.ai/docs/query-offline for more information.

        Parameters
        ----------
        dataset_name
            The name of the `Dataset` to return.
            Previously, you must have supplied a dataset name upon an offline query.
            Dataset names are unique for each environment.
            If 'dataset_name' is provided, then 'job_id' should not be provided.
        dataset_id
            A UUID returned in the `Dataset` object from an offline query.
            Dataset ids are unique for each environment.
            If 'dataset_id' is provided, then 'dataset_name' and 'revision_id' should not be provided.
        revision_id
            The unique id of the `DatasetRevision` to return.
            If a previously-created dataset did not have a name, you can look it
            up using its unique job id instead.
            If 'revision_id' is provided, then 'dataset_name' and 'dataset_id' should not be provided.
        environment
            The environment under which to execute the request.
            API tokens can be scoped to an environment.
            If no environment is specified in the request,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.

        Other Parameters
        ----------------
        job_id
            Same as revision id. Deprecated.

        Returns
        -------
        Dataset
            A `Dataset` that lazily loads your query data.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> uids = [1, 2, 3, 4]
        >>> at = datetime.now(timezone.utc)
        >>> X = ChalkClient().offline_query(
        ...     input={
        ...         User.id: uids,
        ...     },
        ...     input_times=[at] * len(uids),
        ...     output=[
        ...         User.id,
        ...         User.fullname,
        ...         User.email,
        ...         User.name_email_match_score,
        ...     ],
        ...     dataset='my_dataset_name'
        ... )

        Some time later...

        >>> dataset = ChalkClient().get_dataset(
        ...     dataset_name='my_dataset_name'
        ... )
        ...

        or

        >>> dataset = ChalkClient().get_dataset(
        ...     job_id='00000000-0000-0000-0000-000000000000'
        ... )
        ...

        If memory allows:

        >>> df: pd.DataFrame = dataset.get_data_as_pandas()
        """
        ...

    def create_dataset(
        self,
        input: QueryInput,
        dataset_name: str | None = None,
        environment: EnvironmentId | None = None,
        branch: BranchId | None = ...,
        wait: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | ellipsis | None = ...,
    ) -> Dataset:
        """Create a Chalk `Dataset`.

        The `Dataset` wraps a lazily-loading Chalk `DataFrame` that enables us to analyze
        our data without loading all of it directly into memory.
        See https://docs.chalk.ai/docs/query-offline for more information.

        Parameters
        ----------
        input
            The features for which there are known values.
            It can be a mapping of features to a list of values for each
            feature, or an existing `DataFrame`.
        dataset_name
            A unique name that if provided will be used to generate and
            save a `Dataset` constructed from the inputs.
        environment
            The environment under which to execute the request.
            API tokens can be scoped to an environment.
            If no environment is specified in the request,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        wait
            Whether to wait for job completion.
        show_progress
            If True, progress bars will be shown while the query is running.
            Primarily intended for use in a Jupyter-like notebook environment.
            This flag will also be propagated to the methods of the resulting
            `Dataset`.
        timeout:
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        Dataset
            A Chalk `Dataset`.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> uids = [1, 2, 3, 4]
        >>> names = ['a', 'b', 'c', 'd']
        >>> dataset = ChalkClient().create_dataset(
        ...     input={
        ...         User.id: uids,
        ...         User.name: names,
        ...     },
        ...     dataset_name='my_dataset'
        ... )
        >>> df = dataset.get_data_as_pandas()
        """
        ...

    def offline_query(
        self,
        input: Union[QueryInput, OfflineQueryInputUri] | None = None,
        input_times: Sequence[datetime] | datetime | None = None,
        output: Sequence[FeatureReference] = (),
        required_output: Sequence[FeatureReference] = (),
        environment: EnvironmentId | None = None,
        dataset_name: str | None = None,
        branch: BranchId | None = ...,
        correlation_id: str | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
        max_samples: int | None = None,
        wait: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | ellipsis | None = ...,
        recompute_features: bool | list[FeatureReference] = False,
        sample_features: list[FeatureReference] | None = None,
        lower_bound: datetime | timedelta | str | None = None,
        upper_bound: datetime | timedelta | str | None = None,
        store_plan_stages: bool = False,
        explain: bool = False,
        tags: list[str] | None = None,
        required_resolver_tags: list[str] | None = None,
        planner_options: Mapping[str, Union[str, int, bool]] | None = None,
        spine_sql_query: str | None = None,
        resources: ResourceRequests | None = None,
        run_asynchronously: bool = False,
        store_online: bool = False,
        store_offline: bool = False,
        num_shards: int | None = None,
        num_workers: int | None = None,
        completion_deadline: timedelta | None = None,
        max_retries: int | None = None,
    ) -> Dataset:
        """Compute feature values from the offline store or by running offline/online resolvers.
        See `Dataset` for more information.

        Parameters
        ----------
        input
            The features for which there are known values.
            It can be a mapping of features to a list of values for each
            feature, or an existing `DataFrame`.
            Each element in the `DataFrame` or list of values represents
            an observation in line with the timestamp in `input_times`.
        spine_sql_query
            A SQL query that will query your offline store and use the result as input.
            See https://docs.chalk.ai/docs/query-offline#input for more information.
        input_times
            The time at which the given inputs should be observed for point-in-time correctness. If given a list of
            times, the list must match the length of the `input` lists. Each element of input_time corresponds with the
            feature values at the same index of the `input` lists.
            See https://docs.chalk.ai/docs/temporal-consistency for more information.
        output
            The features that you'd like to sample, if they exist.
            If an output feature was never computed for a sample (row) in
            the resulting `DataFrame`, its value will be `None`.
        recompute_features
            Used to control whether resolvers are allowed to run in order to compute feature values.

            If True, all output features will be recomputed by resolvers.
            If False, all output features will be sampled from the offline store.
            If a list, all output features in recompute_features will be recomputed,
            and all other output features will be sampled from the offline store.
        sample_features
            A list of features that will always be sampled, and thus always excluded from recompute.
            Should not overlap with any features used in `recompute_features` argument.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        dataset_name
            A unique name that if provided will be used to generate and
            save a `Dataset` constructed from the list of features computed
            from the inputs.
        max_samples
            The maximum number of samples to include in the `DataFrame`.
            If not specified, all samples will be returned.
        branch
            If specified, Chalk will route your request to the relevant branch.
            If None, Chalk will route your request to a non-branch deployment.
            If not specified, Chalk will use the current client's branch info.
        correlation_id
            You can specify a correlation ID to be used in logs and web interfaces.
            This should be globally unique, i.e. a `uuid` or similar. Logs generated
            during the execution of your query will be tagged with this correlation id.
        query_context
            An immutable context that can be accessed from Python resolvers.
            This context wraps a JSON-compatible dictionary or JSON string with type restrictions.
            See https://docs.chalk.ai/api-docs#ChalkContext for more information.
        wait
            Whether to wait for job completion.
        show_progress
            If True, progress bars will be shown while the query is running.
            Primarily intended for use in a Jupyter-like notebook environment.
            This flag will also be propagated to the methods of the resulting
            `Dataset`.
        timeout:
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.
        lower_bound
            If specified, the query will only be run on data observed after this timestamp.
            Accepts strings in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
        upper_bound
            If specified, the query will only be run on data observed before this timestamp.
            Accepts strings in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
        store_plan_stages
            If True, the output of each of the query plan stages will be stored
            in S3/GCS. This will dramatically impact the performance of the query,
            so it should only be used for debugging.
            These files will be visible in the web dashboard's query detail view, and
            can be downloaded in full by clicking on a plan node in the query plan visualizer.
        tags
            The tags used to scope the resolvers.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        required_resolver_tags
            If specified, *all* required_resolver_tags must be present on a resolver for it to be
            considered eligible to execute.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        resources
            Override resource requests for processes with isolated resources, e.g., offline queries and cron jobs.
            See `ResourceRequests` for more information.
        run_asynchronously
            Boots a kubernetes job to run the queries in their own pods, separate from the engine and branch servers.
            This is useful for large datasets and jobs that require a long time to run.
        store_online
            If True, the output of the query will be stored in the online store.
        store_offline
            If True, the output of the query will be stored in the offline store.
        num_shards
            If specified, the query will be run asynchronously, splitting the input across `num_shards` shards.
        num_workers
            If specified, the query will be run asynchronously across a maximum `num_workers` pod workers at any time.
            This parameter is useful if you have a large number of shards and would like to limit the number of pods running at once.
        completion_deadline
            If specified, shards must complete within 'completion_deadline' duration, or they will be terminated.
            Terminated shards can be tried.
        max_retries
            If specified, failed offline query shards will be retried. The retry budget is shared across all shards.
            By default, max_retries=num_shards/

        Other Parameters
        ----------------
        required_output
            The features that you'd like to sample and must exist
            in each resulting row. Rows where a `required_output`
            was never stored in the offline store will be skipped.
            This differs from specifying the feature in `output`,
            where instead the row would be included, but the feature
            value would be `None`.

        Returns
        -------
        Dataset
            A Chalk `Dataset`.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> uids = [1, 2, 3, 4]
        >>> at = datetime.now(tz=timezone.utc)
        >>> dataset = ChalkClient().offline_query(
        ...     input={
        ...         User.id: uids,
        ...     },
        ...     input_times=[at] * len(uids),
        ...     output=[
        ...         User.id,
        ...         User.fullname,
        ...         User.email,
        ...         User.name_email_match_score,
        ...     ],
        ...     dataset_name='my_dataset'
        ... )
        >>> df = dataset.get_data_as_pandas()
        """
        ...

    def prompt_evaluation(
        self,
        prompts: list[Prompt | str],
        dataset_name: str | None = None,
        dataset_id: str | None = None,
        revision_id: str | None = None,
        reference_output: FeatureReference | None = None,
        evaluators: list[str] | None = None,
        meta: Mapping[str, str] | None = None,
        input: QueryInput | None = None,
        input_times: Sequence[datetime] | datetime | None = None,
        output: Sequence[FeatureReference] = (),
        required_output: Sequence[FeatureReference] = (),
        environment: EnvironmentId | None = None,
        branch: BranchId | None = ...,
        correlation_id: str | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
        max_samples: int | None = None,
        wait: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | ellipsis | None = ...,
        recompute_features: bool | list[FeatureReference] = False,
        sample_features: list[FeatureReference] | None = None,
        lower_bound: datetime | timedelta | str | None = None,
        upper_bound: datetime | timedelta | str | None = None,
        store_plan_stages: bool = False,
        explain: bool = False,
        tags: list[str] | None = None,
        required_resolver_tags: list[str] | None = None,
        planner_options: Mapping[str, Union[str, int, bool]] | None = None,
        spine_sql_query: str | None = None,
        resources: ResourceRequests | None = None,
        run_asynchronously: bool = False,
        store_online: bool = False,
        store_offline: bool = False,
        num_shards: int | None = None,
        num_workers: int | None = None,
        completion_deadline: timedelta | None = None,
        max_retries: int | None = None,
    ) -> Dataset:
        """Runs an evaluation on a set of prompts.
        See https://docs.chalk.ai/docs/prompts#prompt-evaluation for more information.

        Parameters
        ----------
        prompts
            The list of prompts to use for the evaluation.
            This can be a list of `Prompt` objects or a list of named prompts.
        dataset_name
            The name of the `Dataset` to use for the evaluation.
            Dataset names are unique for each environment.
            If 'dataset_name' is provided, then 'dataset_id' should not be provided.
            If 'dataset_name' is provided along with 'inputs', then it will be used to
            generate save a `Dataset` constructed from the list of features computed
            from the inputs.
        dataset_id
            The UUID of the `Dataset` to use for the evaluation.
            Dataset ids are unique for each environment.
            If 'dataset_id' is provided, then 'dataset_name' and 'revision_id' should not be provided.
        revision_id
            The unique id of the `DatasetRevision` to use for the evaluation.
            If a previously-created dataset did not have a name, you can look it
            up using its unique job id instead.
            If 'revision_id' is provided, then 'dataset_name' and 'dataset_id' should not be provided.
        reference_output
            The name of the feature to use as the reference output for the evaluation.
        evaluators
            The list of evaluation functions to use for the evaluation.
            See https://docs.chalk.ai/docs/prompts#prompt-evaluation for more information.
        meta
            Arbitrary `key:value` pairs to associate with a query.
        input
            The features for which there are known values.
            It can be a mapping of features to a list of values for each
            feature, or an existing `DataFrame`.
            Each element in the `DataFrame` or list of values represents
            an observation in line with the timestamp in `input_times`.
        spine_sql_query
            A SQL query that will query your offline store and use the result as input.
            See https://docs.chalk.ai/docs/query-offline#input for more information.
        input_times
            The time at which the given inputs should be observed for point-in-time correctness. If given a list of
            times, the list must match the length of the `input` lists. Each element of input_time corresponds with the
            feature values at the same index of the `input` lists.
            See https://docs.chalk.ai/docs/temporal-consistency for more information.
        output
            The features that you'd like to sample, if they exist.
            If an output feature was never computed for a sample (row) in
            the resulting `DataFrame`, its value will be `None`.
        recompute_features
            Used to control whether resolvers are allowed to run in order to compute feature values.

            If True, all output features will be recomputed by resolvers.
            If False, all output features will be sampled from the offline store.
            If a list, all output features in recompute_features will be recomputed,
            and all other output features will be sampled from the offline store.
        sample_features
            A list of features that will always be sampled, and thus always excluded from recompute.
            Should not overlap with any features used in `recompute_features` argument.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        max_samples
            The maximum number of samples to include in the `DataFrame`.
            If not specified, all samples will be returned.
        branch
            If specified, Chalk will route your request to the relevant branch.
            If None, Chalk will route your request to a non-branch deployment.
            If not specified, Chalk will use the current client's branch info.
        correlation_id
            You can specify a correlation ID to be used in logs and web interfaces.
            This should be globally unique, i.e. a `uuid` or similar. Logs generated
            during the execution of your query will be tagged with this correlation id.
        query_context
            An immutable context that can be accessed from Python resolvers.
            This context wraps a JSON-compatible dictionary or JSON string with type restrictions.
            See https://docs.chalk.ai/api-docs#ChalkContext for more information.
        wait
            Whether to wait for job completion.
        show_progress
            If True, progress bars will be shown while the query is running.
            Primarily intended for use in a Jupyter-like notebook environment.
            This flag will also be propagated to the methods of the resulting
            `Dataset`.
        timeout:
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.
        lower_bound
            If specified, the query will only be run on data observed after this timestamp.
            Accepts strings in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
        upper_bound
            If specified, the query will only be run on data observed before this timestamp.
            Accepts strings in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
        store_plan_stages
            If True, the output of each of the query plan stages will be stored
            in S3/GCS. This will dramatically impact the performance of the query,
            so it should only be used for debugging.
            These files will be visible in the web dashboard's query detail view, and
            can be downloaded in full by clicking on a plan node in the query plan visualizer.
        tags
            The tags used to scope the resolvers.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        required_resolver_tags
            If specified, *all* required_resolver_tags must be present on a resolver for it to be
            considered eligible to execute.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        resources
            Override resource requests for processes with isolated resources, e.g., offline queries and cron jobs.
            See `ResourceRequests` for more information.
        run_asynchronously
            Boots a kubernetes job to run the queries in their own pods, separate from the engine and branch servers.
            This is useful for large datasets and jobs that require a long time to run.
        store_online
            If True, the output of the query will be stored in the online store.
        store_offline
            If True, the output of the query will be stored in the offline store.
        num_shards
            If specified, the query will be run asynchronously, splitting the input across `num_shards` shards.
        num_workers
            If specified, the query will be run asynchronously across a maximum `num_workers` pod workers at any time.
            This parameter is useful if you have a large number of shards and would like to limit the number of pods running at once.
        completion_deadline
            If specified, shards must complete within 'completion_deadline' duration, or they will be terminated.
            Terminated shards can be tried.
        max_retries
            If specified, failed offline query shards will be retried. The retry budget is shared across all shards.
            By default, max_retries=num_shards/

        Other Parameters
        ----------------
        required_output
            The features that you'd like to sample and must exist
            in each resulting row. Rows where a `required_output`
            was never stored in the offline store will be skipped.
            This differs from specifying the feature in `output`,
            where instead the row would be included, but the feature
            value would be `None`.

        Returns
        -------
        Dataset
            A Chalk `Dataset`.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> from chalk.prompts import Prompt, Message
        >>> dataset = ChalkClient().prompt_evaluation(
        ...     dataset_name='my_dataset',
        ...     reference_output='reference_output_column',
        ...     evaluators=['exact_match'],
        ...     prompts=[
        ...         Prompt(model='my_model', messages=[
        ...             Message(role='user', content='what is my name?'),
        ...         ]),
        ...     ]
        ... )
        >>> df = dataset.get_data_as_pandas()
        """
        ...

    def trigger_resolver_run(
        self,
        resolver_fqn: str,
        environment: EnvironmentId | None = None,
        preview_deployment_id: str | None = None,
        branch: BranchId | None = ...,
        upper_bound: datetime | str | None = None,
        lower_bound: datetime | str | None = None,
        store_online: bool = True,
        store_offline: bool = True,
        timestamping_mode: Literal["feature_time", "online_store_write_time"] = "feature_time",
        idempotency_key: Optional[str] = None,
    ) -> ResolverRunResponse:
        """Triggers a resolver to run.
        See https://docs.chalk.ai/docs/runs for more information.

        Parameters
        ----------
        resolver_fqn
            The fully qualified name of the resolver to trigger.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        preview_deployment_id
            If specified, Chalk will route your request to the
            relevant preview deployment.
        upper_bound
            If specified, the resolver will only ingest data observed before this timestamp.
            Accepts strings in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
        lower_bound
            If specified, the resolver will only ingest data observed after this timestamp.
            Accepts strings in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
        store_online
            If True, the resolver run output will be stored in the online store.
        store_offline
            If True, the resolver run output will be stored in the offline store.
        idempotency_key
            If specified, the resolver run will be idempotent with respect to the key.
        branch

        Returns
        -------
        ResolverRunResponse
            Status of the resolver run and the run ID.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> ChalkClient().trigger_resolver_run(
        ...     resolver_fqn="mymodule.fn"
        ... )
        """
        ...

    def get_run_status(
        self,
        run_id: str,
        environment: EnvironmentId | None = None,
        preview_deployment_id: str | None = None,
        branch: BranchId | None = ...,
    ) -> ResolverRunResponse:
        """Retrieves the status of a resolver run.
        See https://docs.chalk.ai/docs/runs for more information.

        Parameters
        ----------
        run_id
            ID of the resolver run to check.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        preview_deployment_id
            If specified, Chalk will route your request to the
            relevant preview deployment.
        branch

        Returns
        -------
        ResolverRunResponse
            Status of the resolver run and the run ID.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> ChalkClient().get_run_status(
        ...     run_id="3",
        ... )
        ResolverRunResponse(
            id="3",
            status=ResolverRunStatus.SUCCEEDED
        )
        """
        ...

    def whoami(self) -> WhoAmIResponse:
        """Checks the identity of your client.

        Useful as a sanity test of your configuration.

        Returns
        -------
        WhoAmIResponse
            The identity of your client.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> ChalkClient().whoami()
        WhoAmIResponse(user="...", environment_id='...', team_id='...')
        """
        ...

    def delete_features(
        self,
        namespace: str,
        features: list[str] | None,
        tags: list[str] | None,
        primary_keys: list[str],
    ) -> FeatureObservationDeletionResponse:
        """Targets feature observation values for deletion and performs deletion online and offline.

        Parameters
        ----------
        namespace
            The namespace in which the target features reside.
        features
            An optional list of the feature names of the features that should be deleted
            for the targeted primary keys. Not specifying this and not specifying the "tags" field
            will result in all features being targeted for deletion for the specified primary keys.
            Note that this parameter and the "tags" parameter are mutually exclusive.
        tags
            An optional list of tags that specify features that should be targeted for deletion.
            If a feature has a tag in this list, its observations for the primary keys you listed
            will be targeted for deletion. Not specifying this and not specifying the "features"
            field will result in all features being targeted for deletion for the specified primary
            keys. Note that this parameter and the "features" parameter are mutually exclusive.
        primary_keys
            The primary keys of the observations that should be targeted for deletion.

        Returns
        -------
        FeatureObservationDeletionResponse
            Holds any errors (if any) that occurred during the drop request.
            Deletion of a feature may partially-succeed.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> ChalkClient().delete_features(
        ...     namespace="user",
        ...     features=["name", "email", "age"],
        ...     primary_keys=[1, 2, 3]
        ... )
        """
        ...

    def drop_features(self, namespace: str, features: list[str]) -> FeatureDropResponse:
        """
        Performs a drop on features, which involves a deletes all their data
        (both online and offline). Once the feature is reset in this manner,
        its type can be changed.

        Parameters
        ----------
        namespace
            The namespace in which the target features reside.
        features
            A list of the feature names of the features that should be dropped.

        Returns
        -------
        FeatureDropResponse
            Holds any errors (if any) that occurred during the drop request.
            Dropping a feature may partially-succeed.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> ChalkClient().drop_features(
        ...     namespace="user",
        ...     features=["name", "email", "age"],
        ... )
        """
        ...

    def upload_features(
        self,
        input: Mapping[FeatureReference, Any],
        branch: BranchId | None = ...,
        environment: EnvironmentId | None = None,
        preview_deployment_id: str | None = None,
        correlation_id: str | None = None,
        query_name: str | None = None,
        meta: Mapping[str, str] | None = None,
    ) -> list[ChalkError] | None:
        """Upload data to Chalk for use in offline resolvers or to prime a cache.

        Parameters
        ----------
        input
            The features for which there are known values, mapped to those values.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        preview_deployment_id
            If specified, Chalk will route your request to the relevant preview deployment
        query_name
            Optionally associate this upload with a query name. See `.query` for more information.

        Other Parameters
        ----------------
        correlation_id
            A globally unique ID for this operation, used alongside logs and
            available in web interfaces.
        meta
            Arbitrary `key:value` pairs to associate with a query.
        branch
            If specified, Chalk will route your request to the relevant branch.

        Returns
        -------
        list[ChalkError] | None
            The errors encountered from uploading features.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> ChalkClient().upload_features(
        ...     input={
        ...         User.id: 1,
        ...         User.name: "Katherine Johnson"
        ...     }
        ... )
        """
        ...

    def multi_upload_features(
        self,
        input: Union[
            list[Mapping[str | Feature | Any, Any]],
            Mapping[str | Feature | Any, list[Any]],
            pd.DataFrame,
            pl.DataFrame,
            DataFrame,
        ],
        branch: BranchId | None = ...,
        environment: EnvironmentId | None = None,
        preview_deployment_id: str | None = None,
        correlation_id: str | None = None,
        meta: Mapping[str, str] | None = None,
    ) -> list[ChalkError] | None:
        """Upload data to Chalk for use in offline resolvers or to prime a cache.

        Parameters
        ----------
        input
            One of three types:
                - A list of mappings, each of which includes the features for which there are known values mapped to
                  those values. Each mapping can have different keys, but each mapping must have the same root features
                  class.
                - A mapping where each feature key is mapped to a list of the values for that feature.
                  You can consider this a mapping that describes columns (keys, i.e. features) and rows
                  (the list of values in the map for each feature). Each list must be the same length.
                - A `pandas`, `polars`, or `chalk.DataFrame`.
        branch
        environment
            The environment under which to run the upload.
            API tokens can be scoped to an environment.
            If no environment is specified in the upload,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        preview_deployment_id
            If specified, Chalk will route your request to the relevant preview deployment

        Other Parameters
        ----------------
        correlation_id
            A globally unique ID for this operation, used alongside logs and
            available in web interfaces. If `None`, a correlation ID will be
            generated for you and returned on the response.
        meta
            Arbitrary `key:value` pairs to associate with an upload.

        Returns
        -------
        list[ChalkError] | None
            The errors encountered from uploading features.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> ChalkClient().multi_upload_features(
        ...     input=[
        ...         {
        ...             User.id: 1,
        ...             User.name: "Katherine Johnson"
        ...         },
        ...         {
        ...             User.id: 2,
        ...             User.name: "Eleanor Roosevelt"
        ...         }
        ...     ]
        ... )
        """
        ...

    def load_features(self, branch: BranchIdParam = ...):
        """Load Chalk features into notebook context. By default, uses the client's
        current branch (if that isn't specified then the main deployment is used).

        Parameters
        ----------
        branch
            If specified, Chalk will route your request to the relevant branch.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> client = ChalkClient(branch='fraud-model')
        ... client.load_features()
        """
        ...

    def sample(
        self,
        output: Sequence[FeatureReference] = (),
        required_output: Sequence[FeatureReference] = (),
        output_id: bool = False,
        output_ts: Union[bool, str] = False,
        max_samples: int | None = None,
        dataset: str | None = None,
        branch: BranchId | None = None,
        environment: EnvironmentId | None = None,
        tags: list[str] | None = None,
    ) -> pd.DataFrame:
        """Get the most recent feature values from the offline store.

        See https://docs.chalk.ai/docs/query-offline for more information.

        Parameters
        ----------
        output
            The features that you'd like to sample, if they exist.
            If an output feature was never computed for a sample (row)
            in the resulting `DataFrame`, its value will be `None`.
        max_samples
            The maximum number of rows to return.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        dataset
            The `Dataset` name under which to save the output.
        tags
            The tags used to scope the resolvers.
            See https://docs.chalk.ai/docs/resolver-tags for more information.

        Other Parameters
        ----------------
        required_output
            The features that you'd like to sample and must exist
            in each resulting row. Rows where a `required_output`
            was never stored in the offline store will be skipped.
            This differs from specifying the feature in `output`,
            where instead the row would be included, but the feature
            value would be `None`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `DataFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `DataFrame`.
        branch

        Returns
        -------
        pd.DataFrame
            A `pandas.DataFrame` with columns equal to the names of the features in output,
            and values representing the value of the most recent observation.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> sample_df = ChalkClient().sample(
        ...     output=[
        ...         Account.id,
        ...         Account.title,
        ...         Account.user.full_name
        ...     ],
        ...     max_samples=10
        ... )
        """
        ...

    def create_branch(
        self,
        branch_name: str,
        create_only: bool = False,
        switch: bool = True,
        source_deployment_id: str | None = None,
        environment: EnvironmentId | None = None,
    ) -> BranchDeployResponse:
        """
        Create a new branch based off of a deployment from the server.
        By default, uses the latest live deployment.

        Parameters
        ----------
        branch_name
            The name of the new branch to create.
        create_only
            If `True`, will raise an error if a branch with the given
            name already exists. If `False` and the branch exists, then
            that branch will be deployed to.
        switch
            If `True`, will switch the client to the newly created branch.
            Defaults to `True`.
        source_deployment_id
            The specific deployment ID to use for the branch.
            If not specified, the latest live deployment on the
            server will be used. You can see which deployments
            are available by clicking on the 'Deployments' tab on
            the project page in the Chalk dashboard.
        environment
            The environment under which to create the branch. API
            tokens can be scoped to an environment. If no environment
            is specified in the query, the environment will be taken
            from the client's cached token.

        Returns
        -------
        BranchDeployResponse
            A response object containing metadata about the branch.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> client = ChalkClient()
        >>> client.create_branch("my-new-branch")
        """
        ...

    def get_or_create_branch(
        self,
        branch_name: str,
        source_branch_name: Optional[str] = None,
        source_deployment_id: Optional[str] = None,
    ):
        """
        Create a new branch named `branch_name` based off of an existing branch or deployment id.
        By default, the latest mainline deployment is used as the branch source.

        If the provided branch name already exists, the client will be updated to point
        to the latest deployment for the already existsing branch (no new deployment
        will be created).

        Parameters
        ----------
        branch_name
            The name to give the newly created branch.
        source_branch_name
            The branch to source the new branch from.
        source_deployment_id
            The specific deployment ID to source the new branch from.
        """

    def get_branches(self) -> list[str]:
        """Lists the current branches for this environment.

        Returns
        -------
        list[str]
            A list of the names of branches available on this environment.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> ChalkClient().get_branches()
        ["testing", "feat/new-feature"]
        """
        ...

    def get_branch(self) -> str | None:
        """Displays the current branch this client is pointed at.

        If the current environment does not support branch deployments
        or no branch is set, this method returns `None`.

        Returns
        -------
        str | None
            The name of the current branch or `None`.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> client = ChalkClient(branch="my-branch")
        >>> assert client.get_branch() == "my-branch"
        """
        ...

    def set_branch(self, branch_name: Optional[str]):
        """Point the `ChalkClient` at the given branch.
        If `branch_name` is None, this points the client at the
        active non-branch deployment.

        If the branch does not exist or if branch deployments
        are not enabled for the current environment, this
        method raises an error.

        Parameters
        ----------
        branch_name
            The name of the branch to use, or None

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> client = ChalkClient()
        >>> client.create_branch("my-new-branch")
        >>> client.set_branch("my-new-branch")
        >>> client.set_branch(None)
        """
        ...

    def reset_branch(self, branch: BranchIdParam = ..., environment: EnvironmentId | None = None):
        ...

    def branch_state(
        self,
        branch: BranchId | ellipsis = ...,
        environment: EnvironmentId | None = None,
    ) -> BranchGraphSummary:
        """
        Returns a `BranchGraphSummary` object that contains the
        state of the branch server: Which resolver/features are
        defined, and the history of live notebook updates on the
        server.

        Parameters
        ----------
        branch
            The branch to query. If not specified, the branch is
            expected to be included in the constructor for `ChalkClient`.
        environment
            Optionally override the environment under which to query the branch state.
        """
        ...

    def set_incremental_cursor(
        self,
        *,
        resolver: str | Resolver | None = None,
        scheduled_query: str | None = None,
        max_ingested_timestamp: datetime | None = None,
        last_execution_timestamp: datetime | None = None,
    ) -> None:
        """
        Sets the incremental cursor for a resolver or scheduled query.

        Parameters
        ---------
        resolver
            The resolver. Can be a function or the string name of a function.
            Exactly one of `resolver` and `scheduled_query` is required.
        scheduled_query
            The name of the scheduled query. Exactly one of `resolver` and `scheduled_query`
            is required.
        max_ingested_timestamp
            Set the maximum timestamp of the data ingested by the resolver.
        last_execution_timestamp
            Override the last execution timestamp of the resolver.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> client = ChalkClient()
        >>> client.set_incremental_cursor(
        ...     resolver="my_resolver",
        ...     max_ingested_timestamp=datetime.now(),
        ... )
        """
        ...

    def get_incremental_cursor(
        self,
        *,
        resolver: str | Resolver | None = None,
        scheduled_query: str | None = None,
    ) -> GetIncrementalProgressResponse:
        """
        Gets the incremental cursor for a resolver or scheduled query.

        Parameters
        ---------
        resolver
            The resolver. Can be a function or the string name of a function.
            Exactly one of `resolver` and `scheduled_query` is required.
        scheduled_query
            If updating incremental status of a resolver in the context of a
            scheduled query, the name of the scheduled query.
            Exactly one of `resolver` and `scheduled_query` is required.

        Returns
        ------
        IncrementalStatus
            An object containing the `max_ingested_timestamp` and `incremental_timestamp`.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> client = ChalkClient()
        >>> client.get_incremental_cursor(resolver="my_resolver")
        """
        ...

    def test_streaming_resolver(
        self,
        resolver: str | Resolver,
        num_messages: int | None = None,
        message_filepath: str | None = None,
        message_keys: list[str | None] | None = None,
        message_bodies: "list[str | bytes | BaseModel] | None" = None,
        message_timestamps: list[str | datetime] | None = None,
        branch: BranchId | ellipsis = ...,
        environment: EnvironmentId | None = None,
        kafka_auto_offset_reset: Optional[Literal["earliest", "latest"]] = "earliest",
    ) -> StreamResolverTestResponse:
        """
        Tests a streaming resolver and its ability to parse and resolve messages.
        See https://docs.chalk.ai/docs/streams for more information.

        Parameters
        ----------
        resolver
            The streaming resolver or its string name.
        num_messages
            The number of messages to digest from the stream source.
            As messages may not be incoming into the stream, this action may time out.
        message_filepath
            A filepath from which test messages will be ingested.
            This file should be newline delimited json as follows:

            >>> {"message_key": "my-key", "message_body": {"field1": "value1", "field2": "value2"}}
            >>> {"message_key": "my-key", "message_body": {"field1": "value1", "field2": "value2"}}

            Each line may optionally contain a timezone string as a value to the key "message_timestamp".
        message_keys
            Alternatively, keys can be supplied in code along with the "test_message_bodies" argument.
            Both arguments must be the same length.
        message_bodies
            Message bodies can be supplied in code as strings, bytes, or Pydantic models along with the "test_message_keys" argument.
            Both arguments must be the same length.
        message_timestamps
            Optionally, timestamps can be provided for each message,

        Other Parameters
        ----------
        branch
            If specified, Chalk will route your request to the relevant branch.
        environment
            The environment under which to create the branch. API
            tokens can be scoped to an environment. If no environment
            is specified in the query, the environment will be taken
            from the client's cached token.
        kafka_auto_offset_reset
            The offset to start reading from when consuming messages from a Kafka source for testing.
            If not specified, the default is "earliest".

        Returns
        -------
        StreamResolverTestResponse
            A simple wrapper around a status and optional error message.
            Inspecting `StreamResolverTestResponse.features` will return the test results, if they exist.
            Otherwise, check `StreamResolverTestResponse.errors` and `StreamResolverTestResponse.message` for errors.

        Examples
        --------
        >>> from chalk.streams import stream, KafkaSource
        >>> from chalk.client import ChalkClient
        >>> from chalk.features import Features, features
        >>> from pydantic import BaseModel
        >>> # This code is an example of a simple streaming feature setup. Define the source
        >>> stream_source=KafkaSource(...)
        >>> # Define the features
        >>> @features(etl_offline_to_online=True, max_staleness="7d")
        >>> class StreamingFeature:
        >>>     id: str
        >>>     user_id: str
        >>>     card_id: str
        >>> # Define the streaming message model
        >>> class StreamingMessage(BaseModel):
        >>>     card_id: str
        >>>     user_id: str
        >>> # Define the mapping resolver
        >>> @stream(source=stream_source)
        >>> def our_stream_resolver(
        >>>     m: StreamingMessage,
        >>> ) -> Features[StreamingFeature.id, StreamingFeature.card_id, StreamingFeature.user_id]:
        >>>    return StreamingFeature(
        >>>        id=f"{m.card_id}-{m.user_id}",
        >>>        card_id=m.card_id,
        >>>        user_id=m.user_id,
        >>>    )
        >>> # Once you have done a `chalk apply`, you can test the streaming resolver with custom messages as follows
        >>> client = ChalkClient()
        >>> keys = ["my_key"] * 10
        >>> messages = [StreamingMessage(card_id="1", user_id=str(i)).json() for i in range(10)]
        >>> resp = client.test_streaming_resolver(
        >>>     resolver="our_stream_resolver",
        >>>     message_keys=keys,
        >>>     message_bodies=messages,
        >>> )
        >>> print(resp.features)
        """
        ...

    def ping_engine(self, num: Optional[int] = None) -> int:
        """
        Ping the engine to check if it is alive.

        Parameters
        ----------
        num
            A random number to send to the engine. If not provided, a random number is generated.
            This number will be returned as the response.

        Returns
        -------
        int
            The number sent to the engine.

        Examples
        --------
        >>> from chalk.client import Chalklient
        >>> client = ChalkClient()
        >>> client.ping_engine(3)
        3
        """
        ...

    def get_operation_feature_statistics(self, operation_id: uuid.UUID) -> FeatureStatisticsResponse:
        ...

    def __new__(cls, *args: Any, **kwargs: Any):
        from chalk.client.client_impl import ChalkAPIClientImpl

        return ChalkAPIClientImpl(*args, **kwargs)


ChalkAPIClientProtocol: TypeAlias = ChalkClient
"""Deprecated. Use `ChalkClient` instead."""



from chalk._gen.chalk.artifacts.v1 import cdc_pb2 as _cdc_pb2
from chalk._gen.chalk.artifacts.v1 import chart_pb2 as _chart_pb2
from chalk._gen.chalk.artifacts.v1 import cron_query_pb2 as _cron_query_pb2
from chalk._gen.chalk.common.v1 import chalk_error_pb2 as _chalk_error_pb2
from chalk._gen.chalk.graph.v1 import graph_pb2 as _graph_pb2
from chalk._gen.chalk.lsp.v1 import lsp_pb2 as _lsp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class ValidationLogSeverity(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    VALIDATION_LOG_SEVERITY_UNSPECIFIED: _ClassVar[ValidationLogSeverity]
    VALIDATION_LOG_SEVERITY_INFO: _ClassVar[ValidationLogSeverity]
    VALIDATION_LOG_SEVERITY_WARNING: _ClassVar[ValidationLogSeverity]
    VALIDATION_LOG_SEVERITY_ERROR: _ClassVar[ValidationLogSeverity]

VALIDATION_LOG_SEVERITY_UNSPECIFIED: ValidationLogSeverity
VALIDATION_LOG_SEVERITY_INFO: ValidationLogSeverity
VALIDATION_LOG_SEVERITY_WARNING: ValidationLogSeverity
VALIDATION_LOG_SEVERITY_ERROR: ValidationLogSeverity

class EnvironmentSettings(_message.Message):
    __slots__ = ("id", "runtime", "requirements", "dockerfile", "requires_packages", "platform_version")
    ID_FIELD_NUMBER: _ClassVar[int]
    RUNTIME_FIELD_NUMBER: _ClassVar[int]
    REQUIREMENTS_FIELD_NUMBER: _ClassVar[int]
    DOCKERFILE_FIELD_NUMBER: _ClassVar[int]
    REQUIRES_PACKAGES_FIELD_NUMBER: _ClassVar[int]
    PLATFORM_VERSION_FIELD_NUMBER: _ClassVar[int]
    id: str
    runtime: str
    requirements: str
    dockerfile: str
    requires_packages: _containers.RepeatedScalarFieldContainer[str]
    platform_version: str
    def __init__(
        self,
        id: _Optional[str] = ...,
        runtime: _Optional[str] = ...,
        requirements: _Optional[str] = ...,
        dockerfile: _Optional[str] = ...,
        requires_packages: _Optional[_Iterable[str]] = ...,
        platform_version: _Optional[str] = ...,
    ) -> None: ...

class ProjectSettings(_message.Message):
    __slots__ = ("project", "environments", "validation")
    PROJECT_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENTS_FIELD_NUMBER: _ClassVar[int]
    VALIDATION_FIELD_NUMBER: _ClassVar[int]
    project: str
    environments: _containers.RepeatedCompositeFieldContainer[EnvironmentSettings]
    validation: ValidationSettings
    def __init__(
        self,
        project: _Optional[str] = ...,
        environments: _Optional[_Iterable[_Union[EnvironmentSettings, _Mapping]]] = ...,
        validation: _Optional[_Union[ValidationSettings, _Mapping]] = ...,
    ) -> None: ...

class MetadataSettings(_message.Message):
    __slots__ = ("name", "missing")
    NAME_FIELD_NUMBER: _ClassVar[int]
    MISSING_FIELD_NUMBER: _ClassVar[int]
    name: str
    missing: str
    def __init__(self, name: _Optional[str] = ..., missing: _Optional[str] = ...) -> None: ...

class FeatureSettings(_message.Message):
    __slots__ = ("metadata",)
    METADATA_FIELD_NUMBER: _ClassVar[int]
    metadata: _containers.RepeatedCompositeFieldContainer[MetadataSettings]
    def __init__(self, metadata: _Optional[_Iterable[_Union[MetadataSettings, _Mapping]]] = ...) -> None: ...

class ResolverSettings(_message.Message):
    __slots__ = ("metadata",)
    METADATA_FIELD_NUMBER: _ClassVar[int]
    metadata: _containers.RepeatedCompositeFieldContainer[MetadataSettings]
    def __init__(self, metadata: _Optional[_Iterable[_Union[MetadataSettings, _Mapping]]] = ...) -> None: ...

class ValidationSettings(_message.Message):
    __slots__ = ("feature", "resolver")
    FEATURE_FIELD_NUMBER: _ClassVar[int]
    RESOLVER_FIELD_NUMBER: _ClassVar[int]
    feature: FeatureSettings
    resolver: ResolverSettings
    def __init__(
        self,
        feature: _Optional[_Union[FeatureSettings, _Mapping]] = ...,
        resolver: _Optional[_Union[ResolverSettings, _Mapping]] = ...,
    ) -> None: ...

class FailedImport(_message.Message):
    __slots__ = ("file_name", "module", "traceback")
    FILE_NAME_FIELD_NUMBER: _ClassVar[int]
    MODULE_FIELD_NUMBER: _ClassVar[int]
    TRACEBACK_FIELD_NUMBER: _ClassVar[int]
    file_name: str
    module: str
    traceback: str
    def __init__(
        self, file_name: _Optional[str] = ..., module: _Optional[str] = ..., traceback: _Optional[str] = ...
    ) -> None: ...

class ChalkpyInfo(_message.Message):
    __slots__ = ("version", "python")
    VERSION_FIELD_NUMBER: _ClassVar[int]
    PYTHON_FIELD_NUMBER: _ClassVar[int]
    version: str
    python: str
    def __init__(self, version: _Optional[str] = ..., python: _Optional[str] = ...) -> None: ...

class ValidationLog(_message.Message):
    __slots__ = ("header", "subheader", "severity")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    SUBHEADER_FIELD_NUMBER: _ClassVar[int]
    SEVERITY_FIELD_NUMBER: _ClassVar[int]
    header: str
    subheader: str
    severity: ValidationLogSeverity
    def __init__(
        self,
        header: _Optional[str] = ...,
        subheader: _Optional[str] = ...,
        severity: _Optional[_Union[ValidationLogSeverity, str]] = ...,
    ) -> None: ...

class Export(_message.Message):
    __slots__ = (
        "graph",
        "crons",
        "charts",
        "cdc_sources",
        "config",
        "chalkpy",
        "failed",
        "logs",
        "lsp",
        "conversion_errors",
    )
    GRAPH_FIELD_NUMBER: _ClassVar[int]
    CRONS_FIELD_NUMBER: _ClassVar[int]
    CHARTS_FIELD_NUMBER: _ClassVar[int]
    CDC_SOURCES_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    CHALKPY_FIELD_NUMBER: _ClassVar[int]
    FAILED_FIELD_NUMBER: _ClassVar[int]
    LOGS_FIELD_NUMBER: _ClassVar[int]
    LSP_FIELD_NUMBER: _ClassVar[int]
    CONVERSION_ERRORS_FIELD_NUMBER: _ClassVar[int]
    graph: _graph_pb2.Graph
    crons: _containers.RepeatedCompositeFieldContainer[_cron_query_pb2.CronQuery]
    charts: _containers.RepeatedCompositeFieldContainer[_chart_pb2.Chart]
    cdc_sources: _containers.RepeatedCompositeFieldContainer[_cdc_pb2.CDCSource]
    config: ProjectSettings
    chalkpy: ChalkpyInfo
    failed: _containers.RepeatedCompositeFieldContainer[FailedImport]
    logs: _containers.RepeatedCompositeFieldContainer[ValidationLog]
    lsp: _lsp_pb2.LSP
    conversion_errors: _containers.RepeatedCompositeFieldContainer[_chalk_error_pb2.ChalkError]
    def __init__(
        self,
        graph: _Optional[_Union[_graph_pb2.Graph, _Mapping]] = ...,
        crons: _Optional[_Iterable[_Union[_cron_query_pb2.CronQuery, _Mapping]]] = ...,
        charts: _Optional[_Iterable[_Union[_chart_pb2.Chart, _Mapping]]] = ...,
        cdc_sources: _Optional[_Iterable[_Union[_cdc_pb2.CDCSource, _Mapping]]] = ...,
        config: _Optional[_Union[ProjectSettings, _Mapping]] = ...,
        chalkpy: _Optional[_Union[ChalkpyInfo, _Mapping]] = ...,
        failed: _Optional[_Iterable[_Union[FailedImport, _Mapping]]] = ...,
        logs: _Optional[_Iterable[_Union[ValidationLog, _Mapping]]] = ...,
        lsp: _Optional[_Union[_lsp_pb2.LSP, _Mapping]] = ...,
        conversion_errors: _Optional[_Iterable[_Union[_chalk_error_pb2.ChalkError, _Mapping]]] = ...,
    ) -> None: ...



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

from chalk._gen.chalk.server.v1 import status_pb2 as chalk_dot_server_dot_v1_dot_status__pb2


class HealthServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CheckHealth = channel.unary_unary(
            "/chalk.server.v1.HealthService/CheckHealth",
            request_serializer=chalk_dot_server_dot_v1_dot_status__pb2.CheckHealthRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_status__pb2.CheckHealthResponse.FromString,
        )
        self.GetHealth = channel.unary_unary(
            "/chalk.server.v1.HealthService/GetHealth",
            request_serializer=chalk_dot_server_dot_v1_dot_status__pb2.GetHealthRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_status__pb2.GetHealthResponse.FromString,
        )
        self.GetClusterMetrics = channel.unary_unary(
            "/chalk.server.v1.HealthService/GetClusterMetrics",
            request_serializer=chalk_dot_server_dot_v1_dot_status__pb2.GetClusterMetricsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_status__pb2.GetClusterMetricsResponse.FromString,
        )


class HealthServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def CheckHealth(self, request, context):
        """If any checks fail, this request fails."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetHealth(self, request, context):
        """Return the actual health check values."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetClusterMetrics(self, request, context):
        """Return collected cluster prometheus metrics"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_HealthServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "CheckHealth": grpc.unary_unary_rpc_method_handler(
            servicer.CheckHealth,
            request_deserializer=chalk_dot_server_dot_v1_dot_status__pb2.CheckHealthRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_status__pb2.CheckHealthResponse.SerializeToString,
        ),
        "GetHealth": grpc.unary_unary_rpc_method_handler(
            servicer.GetHealth,
            request_deserializer=chalk_dot_server_dot_v1_dot_status__pb2.GetHealthRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_status__pb2.GetHealthResponse.SerializeToString,
        ),
        "GetClusterMetrics": grpc.unary_unary_rpc_method_handler(
            servicer.GetClusterMetrics,
            request_deserializer=chalk_dot_server_dot_v1_dot_status__pb2.GetClusterMetricsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_status__pb2.GetClusterMetricsResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("chalk.server.v1.HealthService", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class HealthService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def CheckHealth(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.HealthService/CheckHealth",
            chalk_dot_server_dot_v1_dot_status__pb2.CheckHealthRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_status__pb2.CheckHealthResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetHealth(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.HealthService/GetHealth",
            chalk_dot_server_dot_v1_dot_status__pb2.GetHealthRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_status__pb2.GetHealthResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetClusterMetrics(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.HealthService/GetClusterMetrics",
            chalk_dot_server_dot_v1_dot_status__pb2.GetClusterMetricsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_status__pb2.GetClusterMetricsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )
